from typing import Any, Dict, Tuple, Union

import pandas as pd

from ..registry import NodeRegistry
from ..core.meta.decorators import node_meta
from ..utils import pack_pipeline_output, unpack_pipeline_input
from .base import BaseApplier, BaseCalculator
from ..engines import SkyulfDataFrame, get_engine

# --- Deduplicate ---


class DeduplicateApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        subset = params.get("subset")
        keep = params.get("keep", "first")

        # Handle 'none' string from config
        if keep == "none":
            keep = False

        if subset:
            subset = [c for c in subset if c in X.columns]
            if not subset:
                subset = None

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            X_pl: Any = X
            
            # Map keep parameter
            # Pandas: 'first', 'last', False
            # Polars: 'first', 'last', 'none' (if False)
            pl_keep = keep
            if keep is False:
                pl_keep = "none"
            
            if is_tuple and y is not None:
                # We need to sync X and y
                # Combine them
                # We need to ensure y name doesn't conflict
                # Assuming y is a Series or DataFrame with 1 col
                
                # If y is a Series/DataFrame, we can hstack
                # But we need to know which columns belong to X and which to y
                x_cols = X_pl.columns
                
                # If y is unnamed or has name collision, rename it temporarily?
                # Or just use index? Polars has no index.
                # Best way: add a row index, filter X, get kept indices, filter y.
                
                X_with_idx = X_pl.with_row_index("__idx__")
                X_dedup = X_with_idx.unique(subset=subset, keep=pl_keep, maintain_order=True)
                kept_indices = X_dedup["__idx__"]
                
                # Filter y
                # y must be a DataFrame or Series. If it's a Series, convert to DF to filter?
                # Or use filter/take
                if isinstance(y, pl.DataFrame):
                    y_dedup = y.with_row_index("__idx__").filter(pl.col("__idx__").is_in(kept_indices)).drop("__idx__")
                elif isinstance(y, pl.Series):
                    # Series doesn't have with_row_index directly in same way? 
                    # Actually Series has no index. We can use take/gather.
                    y_dedup = y.gather(kept_indices)
                else:
                    # Should not happen if unpack works correctly
                    y_dedup = y
                
                X_out = X_dedup.drop("__idx__")
                return pack_pipeline_output(X_out, y_dedup, is_tuple)
            
            else:
                X_out = X_pl.unique(subset=subset, keep=pl_keep, maintain_order=True)
                return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        X_dedup = X.drop_duplicates(subset=subset, keep=keep)

        if is_tuple and y is not None:
            # Align y with X
            y_dedup = y.loc[X_dedup.index]
            return pack_pipeline_output(X_dedup, y_dedup, is_tuple)

        return pack_pipeline_output(X_dedup, y, is_tuple)


@NodeRegistry.register("Deduplicate", DeduplicateApplier)
@node_meta(
    id="Deduplicate",
    name="Deduplicate",
    category="Data Operations",
    description="Drop duplicate rows.",
    params={"subset": [], "keep": "first"}
)
class DeduplicateCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        X, _, _ = unpack_pipeline_input(df)

        # Config: {'subset': [...], 'keep': 'first'|'last'|False}
        # Deduplication is an operation that doesn't learn parameters from data,
        # it just applies logic. So fit just passes through the config.

        subset = config.get("subset")
        keep = config.get("keep", "first")

        return {"type": "deduplicate", "subset": subset, "keep": keep}


# --- Drop Missing Columns ---


class DropMissingColumnsApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        cols_to_drop = params.get("columns_to_drop", [])
        
        # Checking X.columns might be tricky if X is Any, but engine check helps
        if engine.name == "polars":
            # X is likely Polars, but let's be safe
            X_pl: Any = X
            cols_to_drop_X = [c for c in cols_to_drop if c in X_pl.columns]
        else:
            cols_to_drop_X = [c for c in cols_to_drop if c in X.columns]


        if cols_to_drop_X:
            if engine.name == "polars":
                X_pl_data: Any = X
                X = X_pl_data.drop(cols_to_drop_X)
            else:
                X = X.drop(columns=cols_to_drop_X)

        return pack_pipeline_output(X, y, is_tuple)


@NodeRegistry.register("DropMissingColumns", DropMissingColumnsApplier)
@node_meta(
    id="DropMissingColumns",
    name="Drop Missing Columns",
    category="Cleaning",
    description="Drop columns that exceed missing value threshold.",
    params={"threshold": 0.5}
)
class DropMissingColumnsCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        X, _, _ = unpack_pipeline_input(df)
        engine = get_engine(X)

        # Config: {'threshold': 50.0 (percent), 'columns': [...]}
        # Threshold is percentage of missing values allowed. If missing > threshold, drop.

        threshold = config.get("missing_threshold")
        explicit_cols = config.get("columns", [])

        cols_to_drop = set()

        # Handle X access depending on engine
        if engine.name == "polars":
            X_pl: Any = X
            if explicit_cols:
                cols_to_drop.update([c for c in explicit_cols if c in X_pl.columns])
        else:
            if explicit_cols:
                cols_to_drop.update([c for c in explicit_cols if c in X.columns])


        if threshold is not None:
            try:
                threshold_val = float(threshold)
                if threshold_val > 0:
                    if engine.name == "polars":
                        import polars as pl
                        X_pl_data: Any = X
                        # Calculate missing percentage for all columns
                        # null_count() returns a DF with 1 row
                        null_counts = X_pl_data.null_count()
                        total_rows = X_pl_data.height
                        
                        for col in X_pl_data.columns:
                            # Get null count for this column
                            # null_counts[col] is a Series of length 1
                            n_null = null_counts[col][0]
                            pct = (n_null / total_rows) * 100
                            if pct >= threshold_val:
                                cols_to_drop.add(col)
                    else:
                        missing_pct = X.isna().mean() * 100
                        auto_dropped = missing_pct[
                            missing_pct >= threshold_val
                        ].index.tolist()
                        cols_to_drop.update(auto_dropped)
            except (TypeError, ValueError):
                pass

        return {
            "type": "drop_missing_columns",
            "columns_to_drop": list(cols_to_drop),
            "threshold": threshold,
        }


# --- Drop Missing Rows ---


class DropMissingRowsApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        subset = params.get("subset")
        how = params.get("how", "any")
        threshold = params.get("threshold")

        if subset:
            if engine.name == "polars":
                X_pl: Any = X
                subset = [c for c in subset if c in X_pl.columns]
            else:
                subset = [c for c in subset if c in X.columns]
            
            if not subset:
                subset = None

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            X_pl_data: Any = X
            
            # We need to sync X and y, so we use the index trick
            X_with_idx = X_pl_data.with_row_index("__idx__")
            
            # Determine columns to check
            check_cols = subset if subset else [c for c in X_pl_data.columns if c != "__idx__"]
            
            if threshold is not None:
                # Keep rows with at least 'threshold' non-null values in check_cols
                # sum_horizontal of is_not_null
                X_clean = X_with_idx.filter(
                    pl.sum_horizontal(pl.col(check_cols).is_not_null()) >= threshold
                )
            elif how == "all":
                # Drop if ALL are null
                # Keep if NOT ALL are null
                X_clean = X_with_idx.filter(
                    ~pl.all_horizontal(pl.col(check_cols).is_null())
                )
            else:
                # how == "any" (default)
                # Drop if ANY is null
                X_clean = X_with_idx.drop_nulls(subset=check_cols)
                
            kept_indices = X_clean["__idx__"]
            
            if is_tuple and y is not None:
                if isinstance(y, pl.DataFrame):
                    y_clean = y.with_row_index("__idx__").filter(pl.col("__idx__").is_in(kept_indices)).drop("__idx__")
                elif isinstance(y, pl.Series):
                    y_clean = y.gather(kept_indices)
                else:
                    y_clean = y
                
                X_out = X_clean.drop("__idx__")
                return pack_pipeline_output(X_out, y_clean, is_tuple)
            
            X_out = X_clean.drop("__idx__")
            return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        # Pandas dropna forbids setting both 'how' and 'thresh'.
        # If 'thresh' is provided (not None), it takes precedence over 'how'.
        if threshold is not None:
            X_clean = X.dropna(axis=0, thresh=threshold, subset=subset)
        else:
            X_clean = X.dropna(axis=0, how=how, subset=subset)

        if is_tuple and y is not None:
            y_clean = y.loc[X_clean.index]
            return pack_pipeline_output(X_clean, y_clean, is_tuple)

        return pack_pipeline_output(X_clean, y, is_tuple)


@NodeRegistry.register("DropMissingRows", DropMissingRowsApplier)
@node_meta(
    id="DropMissingRows",
    name="Drop Missing Rows",
    category="Cleaning",
    description="Drop rows containing missing values in specified columns.",
    params={"subset": [], "how": "any"}
)
class DropMissingRowsCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        # Config: {'subset': [...], 'how': 'any'|'all', 'threshold': int}
        subset = config.get("subset")
        how = config.get("how", "any")
        threshold = config.get("threshold")

        return {
            "type": "drop_missing_rows",
            "subset": subset,
            "how": how,
            "threshold": threshold,
        }


# --- Missing Indicator ---


class MissingIndicatorApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        cols = params.get("columns", [])
        
        if not cols:
            return pack_pipeline_output(X, y, is_tuple)

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            X_pl: Any = X
            
            exprs = []
            for col in cols:
                # X_pl.columns check
                if col in X_pl.columns:
                    exprs.append(pl.col(col).is_null().cast(pl.Int64).alias(f"{col}_missing"))
            
            if not exprs:
                 return pack_pipeline_output(X, y, is_tuple)
                 
            X_out = X_pl.with_columns(exprs)
            return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        X_out = X.copy()
        for col in cols:
            if col in X.columns:
                X_out[f"{col}_missing"] = X[col].isna().astype(int)
                
        return pack_pipeline_output(X_out, y, is_tuple)


@NodeRegistry.register("MissingIndicator", MissingIndicatorApplier)
@node_meta(
    id="MissingIndicator",
    name="Missing Indicator",
    category="Feature Engineering",
    description="Create binary indicators for missing values.",
    params={"features": "missing-only", "sparse": "auto"}
)
class MissingIndicatorCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        X, _, _ = unpack_pipeline_input(df)
        engine = get_engine(X)
        
        explicit_cols = config.get("columns")
        
        if explicit_cols:
            if engine.name == "polars":
                X_pl_data: Any = X
                cols = [c for c in explicit_cols if c in X_pl_data.columns]
            else:
                 cols = [c for c in explicit_cols if c in X.columns]
        else:
            if engine.name == "polars":
                import polars as pl
                X_pl: Any = X
                null_counts = X_pl.null_count()
                cols = [c for c in X_pl.columns if null_counts[c][0] > 0]
            else:
                cols = X.columns[X.isna().any()].tolist()
                
        return {
            "type": "missing_indicator",
            "columns": cols
        }



import polars as pl
import numpy as np
from typing import List, Optional, Tuple
from .schemas import CorrelationMatrix

def calculate_correlations(df: pl.LazyFrame, numeric_cols: List[str]) -> Optional[CorrelationMatrix]:
    """
    Calculates Pearson correlation matrix for numeric columns.
    """
    if len(numeric_cols) < 2:
        return None
        
    try:
        # Polars `corr` is efficient.
        # We need to collect to compute correlation matrix usually.
        # df.select(pl.corr(col1, col2)) for all pairs is one way.
        # Or collect and use .corr() on DataFrame.
        
        # Since correlation matrix is N*N and N is usually small (number of columns),
        # collecting the numeric columns (if not too huge) or computing pair-wise is the way.
        
        # For very large datasets, we should use lazy expressions for covariance and std dev.
        # But `df.corr()` in Polars is optimized.
        
        # Let's try to use the built-in `corr` on the collected DataFrame of just numeric cols.
        # Collecting just numeric cols is usually fine unless we have 10k columns.
        
        # Limit to reasonable number of columns to avoid OOM on huge column sets?
        # Let's assume < 1000 columns for now.
        
        # HARD LIMIT: Top 20 numeric columns to prevent backend/frontend crash
        # Reduced from 50 to 20 as per user report of crashes
        if len(numeric_cols) > 20:
            numeric_cols = numeric_cols[:20]

        subset = df.select(numeric_cols).collect()
        
        # Handle constant columns to avoid RuntimeWarning: invalid value encountered in divide
        # Filter out columns with 0 std dev
        valid_cols = []
        for col in numeric_cols:
            std_val = subset[col].std()
            # Check for None (all nulls or single value) and 0 variance
            if std_val is not None and std_val > 1e-9:
                valid_cols.append(col)
        
        if len(valid_cols) < 2:
            return None
            
        # Re-select only valid columns
        subset = subset.select(valid_cols)
        
        # Suppress numpy warnings that might occur during correlation calculation
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            corr_df = subset.corr()
        
        # Convert to our schema
        matrix = []
        for row in corr_df.iter_rows():
            # Handle NaN/None/Inf
            cleaned_row = []
            for x in row:
                if x is None or np.isnan(x) or np.isinf(x):
                    cleaned_row.append(0.0)
                else:
                    cleaned_row.append(float(x))
            matrix.append(cleaned_row)
            
        return CorrelationMatrix(
            columns=valid_cols,
            values=matrix
        )
        
    except Exception as e:
        print(f"Error calculating correlations: {e}")
        return None



from typing import Any, Dict, List, Optional, Union, TypedDict

class PreprocessingStepConfig(TypedDict, total=False):
    """Configuration for a single preprocessing step."""
    name: str
    transformer: str
    params: Dict[str, Any]

class ModelConfig(TypedDict, total=False):
    """Configuration for the modeling step."""
    type: str # e.g. "random_forest_classifier"
    node_id: str
    params: Dict[str, Any]
    base_model: Dict[str, Any] # nested ModelConfig (recursive type support varies)

class PipelineConfig(TypedDict, total=False):
    """Configuration for the full pipeline."""
    preprocessing: List[PreprocessingStepConfig]
    modeling: ModelConfig

class NodeMetadataDict(TypedDict, total=False):
    """Dictionary representation of node metadata for the registry."""
    id: str
    name: str
    category: str
    description: str
    params: Dict[str, Any]
    tags: List[str]



"""Tests for SkyulfPipeline."""

import os
import tempfile

import pandas as pd
from skyulf.pipeline import SkyulfPipeline


def test_end_to_end_pipeline(sample_classification_data):
    """Test full pipeline execution, saving, and loading."""

    # Define Pipeline Config
    pipeline_config = {
        "preprocessing": [
            {
                "name": "imputer",
                "transformer": "SimpleImputer",
                "params": {"strategy": "mean"},
            },
            {
                "name": "encoder",
                "transformer": "OneHotEncoder",
                "params": {"columns": ["category"]},
            },
            {
                "name": "scaler",
                "transformer": "StandardScaler",
                "params": {"columns": ["feature1", "feature2"]},
            },
        ],
        "modeling": {"type": "logistic_regression", "params": {"C": 1.0}},
    }

    # Initialize
    pipeline = SkyulfPipeline(pipeline_config)

    # Fit
    metrics = pipeline.fit(sample_classification_data, target_column="target")

    assert "preprocessing" in metrics
    assert "modeling" in metrics

    # Predict
    # Create new data (subset)
    new_data = sample_classification_data.drop(columns=["target"]).iloc[:10]
    predictions = pipeline.predict(new_data)

    assert len(predictions) == 10
    assert isinstance(predictions, pd.Series)

    # Save & Load
    with tempfile.NamedTemporaryFile(suffix=".pkl", delete=False) as tmp:
        save_path = tmp.name

    try:
        pipeline.save(save_path)

        loaded_pipeline = SkyulfPipeline.load(save_path)

        # Check if loaded pipeline works
        loaded_preds = loaded_pipeline.predict(new_data)

        pd.testing.assert_series_equal(predictions, loaded_preds)

    finally:
        if os.path.exists(save_path):
            os.remove(save_path)



"""Cross-validation logic for V2 modeling."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union, cast

import numpy as np
import pandas as pd
from sklearn.model_selection import (
    KFold,
    ShuffleSplit,
    StratifiedKFold,
    TimeSeriesSplit,
)

from ..engines import SkyulfDataFrame
from ..engines.sklearn_bridge import SklearnBridge

if TYPE_CHECKING:
    from .base import BaseModelApplier, BaseModelCalculator

from .evaluation.common import sanitize_metrics
from .evaluation.metrics import (
    calculate_classification_metrics,
    calculate_regression_metrics,
)


def _aggregate_metrics(
    fold_metrics: List[Dict[str, float]],
) -> Dict[str, Dict[str, float]]:
    """Aggregates metrics across folds (mean and std)."""
    if not fold_metrics:
        return {}

    keys = fold_metrics[0].keys()
    aggregated = {}

    for key in keys:
        values = [m.get(key, np.nan) for m in fold_metrics]
        # Filter nans
        values = [v for v in values if np.isfinite(v)]

        if values:
            aggregated[key] = {
                "mean": float(np.mean(values)),
                "std": float(np.std(values)),
                "min": float(np.min(values)),
                "max": float(np.max(values)),
            }

    return aggregated


def perform_cross_validation(
    calculator: BaseModelCalculator,
    applier: BaseModelApplier,
    X: Union[pd.DataFrame, SkyulfDataFrame],
    y: Union[pd.Series, Any],
    config: Dict[str, Any],
    n_folds: int = 5,
    cv_type: str = "k_fold",  # k_fold, stratified_k_fold, time_series_split, shuffle_split
    shuffle: bool = True,
    random_state: int = 42,
    progress_callback: Optional[Callable[[int, int], None]] = None,
    log_callback: Optional[Callable[[str], None]] = None,
) -> Dict[str, Any]:
    """
    Performs K-Fold cross-validation.

    Args:
        calculator: The model calculator (fit logic).
        applier: The model applier (predict logic).
        X: Features.
        y: Target.
        config: Model configuration.
        n_folds: Number of folds.
        cv_type: Type of CV.
        shuffle: Whether to shuffle data before splitting (for KFold/Stratified).
        random_state: Random seed for shuffling.
        progress_callback: Optional callback(current_fold, total_folds).
        log_callback: Optional callback for logging messages.

    Returns:
        Dict containing aggregated metrics and per-fold details.
    """

    problem_type = calculator.problem_type

    if log_callback:
        log_callback(f"Starting Cross-Validation (Folds: {n_folds}, Type: {cv_type})")

    # 1. Setup Splitter
    if cv_type == "time_series_split":
        splitter = TimeSeriesSplit(n_splits=n_folds)
    elif cv_type == "shuffle_split":
        splitter = ShuffleSplit(
            n_splits=n_folds, test_size=0.2, random_state=random_state
        )
    elif cv_type == "stratified_k_fold" and problem_type == "classification":
        splitter = StratifiedKFold(
            n_splits=n_folds,
            shuffle=shuffle,
            random_state=random_state if shuffle else None,
        )
    else:
        # Default to KFold
        splitter = KFold(
            n_splits=n_folds,
            shuffle=shuffle,
            random_state=random_state if shuffle else None,
        )

    fold_results = []

    # Ensure numpy for splitting using the Bridge
    X_arr, y_arr = SklearnBridge.to_sklearn((X, y))

    # 2. Iterate Folds
    for fold_idx, (train_idx, val_idx) in enumerate(splitter.split(X_arr, y_arr)):
        if progress_callback:
            progress_callback(fold_idx + 1, n_folds)

        if log_callback:
            log_callback(f"Processing Fold {fold_idx + 1}/{n_folds}...")

        # Split Data
        # We slice the original X/y to preserve their type (Pandas/Polars) for the calculator
        # Polars supports slicing with numpy arrays via __getitem__
        # Pandas supports slicing via iloc
        
        if hasattr(X, "iloc"):
            X_train_fold = X.iloc[train_idx]
            X_val_fold = X.iloc[val_idx]
        else:
            # Polars or other
            X_train_fold = X[train_idx]
            X_val_fold = X[val_idx]
            
        if hasattr(y, "iloc"):
            y_train_fold = y.iloc[train_idx]
            y_val_fold = y.iloc[val_idx]
        else:
            # Polars Series or numpy array
            y_train_fold = y[train_idx]
            y_val_fold = y[val_idx]

        # Fit
        model_artifact = calculator.fit(X_train_fold, y_train_fold, config)

        # Evaluate
        if problem_type == "classification":
            metrics = calculate_classification_metrics(
                model_artifact, X_val_fold, y_val_fold
            )
        else:
            metrics = calculate_regression_metrics(
                model_artifact, X_val_fold, y_val_fold
            )

        if log_callback:
            # Log a key metric for the fold
            key_metric = "accuracy" if problem_type == "classification" else "r2"
            score = metrics.get(key_metric, 0.0)
            log_callback(f"Fold {fold_idx + 1} completed. {key_metric}: {score:.4f}")

        fold_results.append(
            {
                "fold": fold_idx + 1,
                "metrics": sanitize_metrics(metrics),
                # We could store predictions here if needed, but might be too heavy
            }
        )

    # 3. Aggregate
    fold_metrics = [cast(Dict[str, float], r["metrics"]) for r in fold_results]
    aggregated = _aggregate_metrics(fold_metrics)

    if log_callback:
        log_callback(f"Cross-Validation Completed. Aggregated Metrics: {aggregated}")

    return {
        "aggregated_metrics": aggregated,
        "folds": fold_results,
        "cv_config": {
            "n_folds": n_folds,
            "cv_type": cv_type,
            "shuffle": shuffle,
            "random_state": random_state,
        },
    }



from typing import Dict, List, Optional, Union, Any
from pydantic import BaseModel, Field
from datetime import datetime

class NumericStats(BaseModel):
    mean: Optional[float] = None
    median: Optional[float] = None
    std: Optional[float] = None
    variance: Optional[float] = None
    min: Optional[float] = None
    max: Optional[float] = None
    q25: Optional[float] = None
    q75: Optional[float] = None
    skewness: Optional[float] = None
    kurtosis: Optional[float] = None
    zeros_count: Optional[int] = None
    negatives_count: Optional[int] = None
    normality_test: Optional[Dict[str, Any]] = None

class CategoricalStats(BaseModel):
    unique_count: int
    top_k: List[Dict[str, Any]] = Field(default_factory=list)  # [{"value": "A", "count": 10}, ...]
    rare_labels_count: int = 0

class DateStats(BaseModel):
    min_date: Optional[str] = None
    max_date: Optional[str] = None
    duration_days: Optional[float] = None

class TextStats(BaseModel):
    avg_length: Optional[float] = None
    min_length: Optional[int] = None
    max_length: Optional[int] = None
    common_words: List[Dict[str, Any]] = Field(default_factory=list)
    sentiment_distribution: Optional[Dict[str, float]] = None # {"positive": 0.6, "neutral": 0.3, "negative": 0.1}

class HistogramBin(BaseModel):
    start: float
    end: float
    count: int

class NormalityTestResult(BaseModel):
    test_name: str
    statistic: float
    p_value: float
    is_normal: bool

class CausalNode(BaseModel):
    id: str
    label: str

class CausalEdge(BaseModel):
    source: str
    target: str
    type: str # "directed", "undirected", "bidirected"

class CausalGraph(BaseModel):
    nodes: List[CausalNode]
    edges: List[CausalEdge]

class RuleNode(BaseModel):
    id: int
    feature: Optional[str] = None
    threshold: Optional[float] = None
    impurity: float
    samples: int
    value: List[float] # Class distribution
    class_name: Optional[str] = None # Predicted class
    is_leaf: bool
    children: List[int] = Field(default_factory=list) # IDs of children

class RuleTree(BaseModel):
    nodes: List[RuleNode]
    accuracy: Optional[float] = None # Surrogate model accuracy
    rules: Optional[List[str]] = None # Human readable rules
    feature_importances: Optional[List[Dict[str, Union[str, float]]]] = None # Feature importance from surrogate model

class ColumnProfile(BaseModel):
    name: str
    dtype: str  # "Numeric", "Categorical", "Boolean", "DateTime", "Text"
    missing_count: int
    missing_percentage: float
    
    # Type-specific stats
    numeric_stats: Optional[NumericStats] = None
    categorical_stats: Optional[CategoricalStats] = None
    date_stats: Optional[DateStats] = None
    text_stats: Optional[TextStats] = None
    
    # Distribution
    histogram: Optional[List[HistogramBin]] = None
    normality_test: Optional[NormalityTestResult] = None
    
    # Quality
    is_constant: bool = False
    is_unique: bool = False  # Possible ID

class CorrelationMatrix(BaseModel):
    columns: List[str]
    values: List[List[float]]  # 2D array

class ScatterSample(BaseModel):
    x: str
    y: str
    data: List[Dict[str, Any]]  # [{"x": 1, "y": 2}, ...]

class Alert(BaseModel):
    column: Optional[str] = None
    type: str  # "High Null", "Constant", "High Cardinality", "Leakage", "Outlier"
    message: str
    severity: str = "warning"  # "info", "warning", "error"

class Recommendation(BaseModel):
    column: Optional[str] = None
    action: str # "Drop", "Impute", "Transform", "Encode"
    reason: str
    suggestion: str

class PCAComponent(BaseModel):
    component: str # "PC1", "PC2", "PC3"
    explained_variance_ratio: float
    top_features: Dict[str, float] # feature_name -> weight/loading

class PCAPoint(BaseModel):
    x: float
    y: float
    z: Optional[float] = None
    label: Optional[str] = None # For target coloring

class GeoPoint(BaseModel):
    lat: float
    lon: float
    label: Optional[str] = None

class GeospatialStats(BaseModel):
    lat_col: str
    lon_col: str
    min_lat: float
    max_lat: float
    min_lon: float
    max_lon: float
    centroid_lat: float
    centroid_lon: float
    sample_points: List[GeoPoint]

class TimeSeriesPoint(BaseModel):
    date: str
    values: Dict[str, float]

class BoxPlotStats(BaseModel):
    min: float
    q1: float
    median: float
    q3: float
    max: float

class CategoryBoxPlot(BaseModel):
    name: str
    stats: BoxPlotStats

class TargetInteraction(BaseModel):
    feature: str
    plot_type: str # "boxplot"
    data: List[CategoryBoxPlot]
    p_value: Optional[float] = None # ANOVA p-value

class SeasonalityStats(BaseModel):
    day_of_week: List[Dict[str, Any]]
    month_of_year: List[Dict[str, Any]]

class TimeSeriesAnalysis(BaseModel):
    date_col: str
    trend: List[TimeSeriesPoint]
    seasonality: SeasonalityStats
    autocorrelation: Optional[List[Dict[str, Any]]] = None
    stationarity_test: Optional[Dict[str, Any]] = None

class OutlierPoint(BaseModel):
    index: int
    values: Dict[str, Any] # Key values for context
    score: float # Anomaly score (lower is more anomalous for IF, or distance for others)
    explanation: Optional[List[Dict[str, Any]]] = None # [{"feature": "Age", "value": 95, "mean": 35, "diff": 60}, ...]

class OutlierAnalysis(BaseModel):
    method: str # "IsolationForest" or "IQR"
    total_outliers: int
    outlier_percentage: float
    top_outliers: List[OutlierPoint]
    plot_data: Optional[List[Dict[str, Any]]] = None # For visualization (e.g. PCA projection of outliers)

class ClusteringPoint(BaseModel):
    x: float
    y: float
    cluster: int
    label: Optional[str] = None

class ClusterStats(BaseModel):
    cluster_id: int
    size: int
    percentage: float
    center: Dict[str, float]

class ClusteringAnalysis(BaseModel):
    method: str = "KMeans"
    n_clusters: int
    inertia: float
    clusters: List[ClusterStats]
    points: List[ClusteringPoint]

class Filter(BaseModel):
    column: str
    operator: str # "==", "!=", ">", "<", ">=", "<=", "in"
    value: Any

class DatasetProfile(BaseModel):
    row_count: int
    column_count: int
    duplicate_rows: int
    missing_cells_percentage: float
    memory_usage_mb: float
    
    columns: Dict[str, ColumnProfile]
    correlations: Optional[CorrelationMatrix] = None
    correlations_with_target: Optional[CorrelationMatrix] = None
    alerts: List[Alert] = Field(default_factory=list)
    recommendations: List[Recommendation] = Field(default_factory=list)
    sample_data: Optional[List[Dict[str, Any]]] = None
    
    # Target Analysis
    target_col: Optional[str] = None
    task_type: Optional[str] = None # "Classification" or "Regression"
    target_correlations: Optional[Dict[str, float]] = None
    target_interactions: Optional[List[TargetInteraction]] = None
    
    # Multivariate
    pca_data: Optional[List[PCAPoint]] = None
    pca_components: Optional[List[PCAComponent]] = None
    outliers: Optional[OutlierAnalysis] = None
    clustering: Optional[ClusteringAnalysis] = None
    causal_graph: Optional[CausalGraph] = None
    rule_tree: Optional[RuleTree] = None
    vif: Optional[Dict[str, float]] = None # Variance Inflation Factor for numeric columns
    
    # Geospatial
    geospatial: Optional[GeospatialStats] = None
    
    # Time Series
    timeseries: Optional[TimeSeriesAnalysis] = None
    
    # Metadata
    excluded_columns: List[str] = Field(default_factory=list)
    active_filters: Optional[List[Filter]] = None
    
    generated_at: datetime = Field(default_factory=datetime.now)



"""Tests for modeling module."""

import pandas as pd
from skyulf.data.dataset import SplitDataset
from skyulf.modeling.base import StatefulEstimator
from skyulf.modeling.classification import (
    LogisticRegressionApplier,
    LogisticRegressionCalculator,
)
from skyulf.modeling.regression import (
    RandomForestRegressorApplier,
    RandomForestRegressorCalculator,
)


def test_logistic_regression_training(sample_classification_data):
    """Test Logistic Regression training and prediction."""
    # Prepare data
    # Drop missing for this test or fill them
    data = sample_classification_data.fillna(0)
    # Drop non-numeric for simple test
    data = data.drop(columns=["category"])

    dataset = SplitDataset(train=data.iloc[:80], test=data.iloc[80:], validation=None)

    estimator = StatefulEstimator(
        node_id="test_lr",
        calculator=LogisticRegressionCalculator(),
        applier=LogisticRegressionApplier(),
    )

    config = {"params": {"C": 0.5}}

    # Fit Predict
    predictions = estimator.fit_predict(dataset, target_column="target", config=config)

    assert "train" in predictions
    assert "test" in predictions
    assert len(predictions["train"]) == 80
    assert len(predictions["test"]) == 20

    # Evaluate
    report = estimator.evaluate(dataset, target_column="target")
    assert report["problem_type"] == "classification"
    assert "accuracy" in report["splits"]["test"].metrics


def test_random_forest_regression(sample_regression_data):
    """Test Random Forest Regression."""
    data = sample_regression_data.drop(columns=["category"])

    dataset = SplitDataset(train=data.iloc[:80], test=data.iloc[80:], validation=None)

    estimator = StatefulEstimator(
        node_id="test_rf",
        calculator=RandomForestRegressorCalculator(),
        applier=RandomForestRegressorApplier(),
    )

    config = {"params": {"n_estimators": 10}}

    estimator.fit_predict(dataset, target_column="target", config=config)

    report = estimator.evaluate(dataset, target_column="target")
    assert report["problem_type"] == "regression"
    assert "mse" in report["splits"]["test"].metrics


def test_cross_validation(sample_classification_data):
    """Test Cross Validation."""
    data = sample_classification_data.fillna(0).drop(columns=["category"])

    dataset = SplitDataset(train=data, test=pd.DataFrame())

    estimator = StatefulEstimator(
        node_id="test_cv",
        calculator=LogisticRegressionCalculator(),
        applier=LogisticRegressionApplier(),
    )

    cv_results = estimator.cross_validate(
        dataset=dataset, target_column="target", config={}, n_folds=3
    )

    assert "aggregated_metrics" in cv_results
    assert "accuracy" in cv_results["aggregated_metrics"]
    assert len(cv_results["folds"]) == 3



"""Main Skyulf Pipeline."""

import logging
import pickle
from typing import Any, Dict, Optional, Union, cast

import pandas as pd

from .types import PipelineConfig
from .data.dataset import SplitDataset
from .engines import SkyulfDataFrame, get_engine
from .modeling.base import BaseModelApplier, BaseModelCalculator, StatefulEstimator
from .modeling.classification import (
    LogisticRegressionApplier,
    LogisticRegressionCalculator,
    RandomForestClassifierApplier,
    RandomForestClassifierCalculator,
)
from .modeling.regression import (
    RandomForestRegressorApplier,
    RandomForestRegressorCalculator,
    RidgeRegressionApplier,
    RidgeRegressionCalculator,
)
from .modeling.tuning.engine import TuningApplier, TuningCalculator
from .preprocessing.pipeline import FeatureEngineer
from .registry import NodeRegistry

logger = logging.getLogger(__name__)


class SkyulfPipeline:
    """
    End-to-end ML Pipeline.

    Encapsulates:
    1. Feature Engineering (Preprocessing)
    2. Modeling (Training/Inference)
    """

    def __init__(self, config: PipelineConfig):
        """
        Initialize the pipeline.

        Args:
            config: Pipeline configuration dictionary.
                    Must contain 'preprocessing' (list) and 'modeling' (dict).
        """
        self.config = config
        self.preprocessing_steps = config.get("preprocessing", [])
        self.modeling_config = config.get("modeling", {})

        self.feature_engineer = FeatureEngineer(self.preprocessing_steps)
        self.model_estimator: Optional[StatefulEstimator] = None

        # Initialize model estimator if config is present
        if self.modeling_config:
            self._init_model_estimator()

    def _init_model_estimator(self):
        """Initialize the StatefulEstimator based on config."""
        model_type = self.modeling_config.get("type")
        if not model_type:
            return

        node_id = self.modeling_config.get("node_id", "model_node")

        calculator: Optional[BaseModelCalculator] = None
        applier: Optional[BaseModelApplier] = None

        # Try Registry first
        if model_type:
            try:
                calculator = NodeRegistry.get_calculator(model_type)()
                applier = NodeRegistry.get_applier(model_type)()
            except ValueError:
                pass

        if calculator is None:
            # Map model types to classes
            if model_type == "logistic_regression":
                calculator = LogisticRegressionCalculator()
                applier = LogisticRegressionApplier()
            elif model_type == "random_forest_classifier":
                calculator = RandomForestClassifierCalculator()
                applier = RandomForestClassifierApplier()
            elif model_type == "ridge_regression":
                calculator = RidgeRegressionCalculator()
                applier = RidgeRegressionApplier()
            elif model_type == "random_forest_regressor":
                calculator = RandomForestRegressorCalculator()
                applier = RandomForestRegressorApplier()
            elif model_type == "hyperparameter_tuner":
                # Tuner wraps another model
                base_model_config = self.modeling_config.get("base_model", {})
                base_model_type = base_model_config.get("type")

                base_calc: Optional[BaseModelCalculator] = None
                base_applier: Optional[BaseModelApplier] = None

                # Try Registry for base model
                if base_model_type:
                    try:
                        base_calc = NodeRegistry.get_calculator(base_model_type)()
                        base_applier = NodeRegistry.get_applier(base_model_type)()
                    except ValueError:
                        pass

                if base_calc is None:
                    if base_model_type == "logistic_regression":
                        base_calc = LogisticRegressionCalculator()
                        base_applier = LogisticRegressionApplier()
                    elif base_model_type == "random_forest_classifier":
                        base_calc = RandomForestClassifierCalculator()
                        base_applier = RandomForestClassifierApplier()
                    elif base_model_type == "ridge_regression":
                        base_calc = RidgeRegressionCalculator()
                        base_applier = RidgeRegressionApplier()
                    elif base_model_type == "random_forest_regressor":
                        base_calc = RandomForestRegressorCalculator()
                        base_applier = RandomForestRegressorApplier()

                if base_calc and base_applier:
                    calculator = TuningCalculator(base_calc)
                    applier = TuningApplier(base_applier)
                else:
                    raise ValueError(
                        f"Unknown base model type for tuner: {base_model_type}"
                    )

        if calculator is None or applier is None:
            raise ValueError(f"Unknown model type: {model_type}")

        self.model_estimator = StatefulEstimator(
            node_id=node_id, calculator=calculator, applier=applier
        )

    def fit(
        self, data: Union[pd.DataFrame, SkyulfDataFrame, SplitDataset], target_column: str
    ) -> Dict[str, Any]:
        """
        Fit the pipeline.

        Args:
            data: Input data (DataFrame or SplitDataset).
            target_column: Name of the target column.

        Returns:
            Dictionary containing execution metrics.
        """
        metrics = {}

        # 1. Feature Engineering
        logger.info("Starting Feature Engineering...")
        transformed_data, fe_metrics = self.feature_engineer.fit_transform(data)
        metrics["preprocessing"] = fe_metrics

        # 2. Modeling
        if self.model_estimator:
            logger.info("Starting Model Training...")

            # Ensure transformed_data is SplitDataset for modeling
            if isinstance(transformed_data, SplitDataset):
                dataset = transformed_data
            else:
                # If we only have a DataFrame, we can't really evaluate properly without a split
                # But we can fit on it.
                # Ideally, the user should provide a SplitDataset or use a Splitter node in preprocessing.
                # If preprocessing didn't split, we wrap it.
                engine = get_engine(transformed_data)
                empty_df = engine.create_dataframe({})
                dataset = SplitDataset(
                    train=transformed_data, test=empty_df, validation=None
                )

            # Fit the model
            # Note: fit_predict updates self.model_estimator.model in-memory
            _ = self.model_estimator.fit_predict(
                dataset=dataset,
                target_column=target_column,
                config=cast(Dict[str, Any], self.modeling_config),
            )

            # Evaluate
            # We can run evaluation if we have test/validation sets
            try:
                eval_report = self.model_estimator.evaluate(
                    dataset=dataset, target_column=target_column
                )
                metrics["modeling"] = eval_report
            except Exception as e:
                logger.warning(f"Evaluation failed: {e}")
                metrics["modeling_error"] = str(e)

        return metrics

    def predict(self, data: Union[pd.DataFrame, SkyulfDataFrame]) -> Any:
        """
        Generate predictions.

        Args:
            data: Input DataFrame.

        Returns:
            Series of predictions.
        """
        # 1. Feature Engineering (Transform only)
        transformed_data = self.feature_engineer.transform(data)

        # 2. Modeling
        if self.model_estimator and self.model_estimator.model is not None:
            return self.model_estimator.applier.predict(
                transformed_data, self.model_estimator.model
            )
        else:
            raise ValueError("Pipeline not fitted or no model configured.")

    def save(self, path: str):
        """Save the pipeline to a file."""
        # We can use pickle to save the whole object since we removed external dependencies
        with open(path, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, path: str) -> "SkyulfPipeline":
        """Load the pipeline from a file."""
        with open(path, "rb") as f:
            return pickle.load(f)  # type: ignore



import logging
from typing import Any, Dict, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.preprocessing import PowerTransformer, StandardScaler

from ..core.meta.decorators import node_meta
from ..registry import NodeRegistry
from ..utils import (
    detect_numeric_columns,
    pack_pipeline_output,
    resolve_columns,
    unpack_pipeline_input,
)
from .base import BaseApplier, BaseCalculator
from ..engines import SkyulfDataFrame, get_engine

logger = logging.getLogger(__name__)

# --- Power Transformer (Box-Cox, Yeo-Johnson) ---


class PowerTransformerApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)
        was_polars = engine.name == "polars"

        cols = params.get("columns", [])
        lambdas = params.get("lambdas")
        method = params.get("method", "yeo-johnson")
        standardize = params.get("standardize", True)
        scaler_params = params.get("scaler_params", {})

        valid_cols = [c for c in cols if c in X.columns]
        if not valid_cols or lambdas is None:
            return pack_pipeline_output(X, y, is_tuple)

        if was_polars:
            import polars as pl
            X_pd = X.to_pandas()
        else:
            X_pd = X

        df_out = X_pd.copy()

        # Reconstruct PowerTransformer state for application
        # We manually restore the lambdas and internal scaler to apply the transform
        # without re-fitting.

        X_vals = df_out[valid_cols].values
        # Filter lambdas and scaler params to match valid_cols
        col_indices = [cols.index(c) for c in valid_cols]
        lambdas_arr = np.array(lambdas)[col_indices]

        # 1. Power Transform

        try:
            pt = PowerTransformer(method=method, standardize=standardize)
            pt.lambdas_ = lambdas_arr

            if standardize:
                scaler = StandardScaler()

                mean = np.array(scaler_params.get("mean"))
                scale = np.array(scaler_params.get("scale"))

                if len(mean) == len(cols):
                    mean = mean[col_indices]
                if len(scale) == len(cols):
                    scale = scale[col_indices]

                scaler.mean_ = mean
                scaler.scale_ = scale
                scaler.var_ = np.square(scaler.scale_)  # Approximate if not stored
                pt._scaler = scaler

            # We need to trick sklearn into thinking it's fitted
            # Usually setting attributes is enough, but let's see.
            # PowerTransformer checks hasattr(self, "lambdas_")

            X_trans = pt.transform(X_vals)
            # sklearn can be configured with transform_output="pandas", which returns a DataFrame.
            X_trans_arr = X_trans.to_numpy() if hasattr(X_trans, "to_numpy") else X_trans
            df_out.loc[:, valid_cols] = np.asarray(X_trans_arr)

        except Exception as e:
            logger.error(f"PowerTransformer application failed: {e}")
            # Fallback?
            pass

        if was_polars:
            df_out = pl.from_pandas(df_out)

        return pack_pipeline_output(df_out, y, is_tuple)


@NodeRegistry.register("PowerTransformer", PowerTransformerApplier)
@node_meta(
    id="PowerTransformer",
    name="Power Transformer",
    category="Preprocessing",
    description="Apply a power transform featurewise to make data more Gaussian-like.",
    params={"method": "yeo-johnson", "standardize": True, "columns": []}
)
class PowerTransformerCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        X, _, _ = unpack_pipeline_input(df)
        engine = get_engine(X)
        if engine.name == "polars":
            X = X.to_pandas()

        # Config: {'method': 'yeo-johnson' | 'box-cox', 'standardize': True, 'columns': [...]}
        method = config.get("method", "yeo-johnson")
        standardize = config.get("standardize", True)

        cols = resolve_columns(X, config, detect_numeric_columns)

        if not cols:
            return {}

        valid_cols = []
        if method == "box-cox":
            for col in cols:
                # Box-Cox requires strictly positive data
                if (X[col] <= 0).any():
                    continue
                valid_cols.append(col)
        else:
            valid_cols = cols

        if not valid_cols:
            return {}

        transformer = PowerTransformer(method=method, standardize=standardize)
        transformer.fit(X[valid_cols])

        # Capture internal scaler parameters if standardization is enabled
        scaler_params = {}
        if standardize and hasattr(transformer, "_scaler"):
            scaler = transformer._scaler
            if scaler:
                scaler_params = {
                    "mean": scaler.mean_.tolist() if scaler.mean_ is not None else None,
                    "scale": (
                        scaler.scale_.tolist() if scaler.scale_ is not None else None
                    ),
                }

        return {
            "type": "power_transformer",
            "lambdas": transformer.lambdas_.tolist(),
            "method": method,
            "standardize": standardize,
            "columns": valid_cols,
            "scaler_params": scaler_params,
        }


# --- Simple Transformations (Log, Sqrt, etc.) ---


class SimpleTransformationApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        transformations = params.get("transformations", [])
        if not transformations:
            return pack_pipeline_output(X, y, is_tuple)

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            X_pl: Any = X
            X_out = X_pl

            for item in transformations:
                col = item.get("column")
                method = item.get("method")
                if col not in X_out.columns:
                    continue
                
                expr = pl.col(col)
                
                if method == "log":
                    expr = pl.when(pl.col(col) < 0).then(None).otherwise(pl.col(col)).log1p()
                elif method == "square_root":
                    # sqrt of neg is nan
                    expr = pl.when(pl.col(col) < 0).then(None).otherwise(pl.col(col)).sqrt()
                elif method == "cube_root":
                    expr = pl.col(col).cbrt()
                elif method == "reciprocal":
                    # 1/0 is inf in polars, typically nan in pandas logic above (replace(0, nan))
                    expr = 1.0 / pl.when(pl.col(col) == 0).then(None).otherwise(pl.col(col))
                elif method == "square":
                    expr = pl.col(col).pow(2)
                elif method == "exponential":
                    threshold = item.get("clip_threshold", 700)
                    expr = pl.col(col).clip(upper_bound=threshold).exp()
                
                X_out = X_out.with_columns(expr.alias(col))
            
            return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        df_out = X.copy()

        for item in transformations:
            col = item.get("column")
            method = item.get("method")

            if col not in df_out.columns:
                continue

            series = pd.to_numeric(df_out[col], errors="coerce")

            if method == "log":
                # log1p is safer for zeros
                if (series < 0).any():
                    series[series < 0] = np.nan
                df_out[col] = np.log1p(series)

            elif method == "square_root":
                if (series < 0).any():
                    series[series < 0] = np.nan
                df_out[col] = np.sqrt(series)

            elif method == "cube_root":
                df_out[col] = np.cbrt(series)

            elif method == "reciprocal":
                df_out[col] = 1.0 / series.replace(0, np.nan)

            elif method == "square":
                df_out[col] = np.square(series)

            elif method == "exponential":
                # Clip to avoid overflow (exp(709) ~ max float64)
                # We use a slightly lower bound to be safe, or user provided threshold
                threshold = item.get("clip_threshold", 700)
                series_clipped = series.clip(upper=threshold)
                df_out[col] = np.exp(series_clipped)

        return pack_pipeline_output(df_out, y, is_tuple)


@NodeRegistry.register("SimpleTransformation", SimpleTransformationApplier)
@node_meta(
    id="SimpleTransformation",
    name="Simple Transformation",
    category="Preprocessing",
    description="Apply simple mathematical transformations (log, sqrt, etc.).",
    params={"func": "log", "columns": []}
)
class SimpleTransformationCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Dict[str, Any]:
        # Config: {'transformations': [{'column': 'col1', 'method': 'log'}, ...]}
        return {
            "type": "simple_transformation",
            "transformations": params.get("transformations", []),
        }


# --- General Transformation (Combined) ---


class GeneralTransformationApplier(BaseApplier):
    def apply(  # noqa: C901
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        transformations = params.get("transformations", [])
        if not transformations:
            return pack_pipeline_output(X, y, is_tuple)
        
        # Polars Path
        if engine.name == "polars":
            import polars as pl
            X_pl: Any = X
            X_out = X_pl
            
            for item in transformations:
                col = item.get("column")
                method = item.get("method")
                
                if col not in X_out.columns:
                    continue

                if method in ["box-cox", "yeo-johnson"]:
                    lambdas = item.get("lambdas")
                    scaler_params = item.get("scaler_params")
                    if lambdas is None:
                        continue
                    
                    try:
                        pt = PowerTransformer(method=method, standardize=True)
                        pt.lambdas_ = np.array(lambdas)
                        
                        if scaler_params:
                            scaler = StandardScaler()
                            # Handle potential None or list
                            m = scaler_params.get("mean")
                            s = scaler_params.get("scale")
                            if m is not None:
                                scaler.mean_ = np.array(m)
                            if s is not None:
                                scaler.scale_ = np.array(s)
                                scaler.var_ = np.square(scaler.scale_)
                            pt._scaler = scaler
                        
                        # Get numpy array from polars col
                        vals = X_out[col].to_numpy().reshape(-1, 1)
                        trans_vals = pt.transform(vals)
                        # flatten
                        flat_vals = trans_vals.ravel()
                        
                        X_out = X_out.with_columns(pl.Series(flat_vals).alias(col))
                        
                    except Exception as e:
                        logger.warning(f"Failed to apply {method} for column {col}: {e}")
                
                # Simple Transformations (Polars native)
                else: 
                    expr = pl.col(col)
                    if method == "log":
                        expr = pl.when(pl.col(col) < 0).then(None).otherwise(pl.col(col)).log1p()
                    elif method in ["sqrt", "square_root"]:
                        expr = pl.when(pl.col(col) < 0).then(None).otherwise(pl.col(col)).sqrt()
                    elif method == "cube_root":
                        expr = pl.col(col).cbrt()
                    elif method == "reciprocal":
                        expr = 1.0 / pl.when(pl.col(col) == 0).then(None).otherwise(pl.col(col))
                    elif method == "square":
                        expr = pl.col(col).pow(2)
                    elif method in ["exp", "exponential"]:
                        threshold = item.get("clip_threshold", 700)
                        expr = pl.col(col).clip(upper_bound=threshold).exp()
                    
                    X_out = X_out.with_columns(expr.alias(col))
            
            return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        df_out = X.copy()

        for item in transformations:
            col = item.get("column")
            method = item.get("method")

            if col not in df_out.columns:
                continue

            series = pd.to_numeric(df_out[col], errors="coerce")

            if method in ["box-cox", "yeo-johnson"]:
                lambdas = item.get("lambdas")
                scaler_params = item.get("scaler_params")

                if lambdas is None:
                    continue

                try:
                    pt = PowerTransformer(method=method, standardize=True)
                    pt.lambdas_ = np.array(lambdas)

                    if scaler_params:
                        scaler = StandardScaler()
                        scaler.mean_ = np.array(scaler_params.get("mean"))
                        scaler.scale_ = np.array(scaler_params.get("scale"))
                        scaler.var_ = np.square(scaler.scale_)
                        pt._scaler = scaler

                    # Reshape for sklearn
                    vals = series.values.reshape(-1, 1)
                    trans_vals = pt.transform(vals)
                    # sklearn can be configured with transform_output="pandas", which returns a DataFrame.
                    trans_vals_arr = (
                        trans_vals.to_numpy() if hasattr(trans_vals, "to_numpy") else trans_vals
                    )
                    df_out[col] = np.asarray(trans_vals_arr).ravel()
                except Exception as e:
                    logger.warning(f"Failed to apply {method} for column {col}: {e}")

            elif method == "log":
                if (series < 0).any():
                    series[series < 0] = np.nan
                df_out[col] = np.log1p(series)
            elif method == "sqrt" or method == "square_root":
                if (series < 0).any():
                    series[series < 0] = np.nan
                df_out[col] = np.sqrt(series)
            elif method == "cube_root":
                df_out[col] = np.cbrt(series)
            elif method == "reciprocal":
                df_out[col] = 1.0 / series.replace(0, np.nan)
            elif method == "square":
                df_out[col] = np.square(series)
            elif method == "exp" or method == "exponential":
                threshold = item.get("clip_threshold", 700)
                series_clipped = series.clip(upper=threshold)
                df_out[col] = np.exp(series_clipped)

        return pack_pipeline_output(df_out, y, is_tuple)


@NodeRegistry.register("GeneralTransformation", GeneralTransformationApplier)
@node_meta(
    id="GeneralTransformation",
    name="General Transformation",
    category="Preprocessing",
    description="Apply various function transformations (log, sqrt, square, exp) to columns.",
    params={"transformations": []}
)
class GeneralTransformationCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        # Config: {'transformations': [{'column': 'col1', 'method': 'log'},
        #                              {'column': 'col2', 'method': 'yeo-johnson'}]}
        X, _, _ = unpack_pipeline_input(df)
        engine = get_engine(X)

        transformations_config = config.get("transformations", [])
        fitted_transformations = []

        for item in transformations_config:
            col = item.get("column")
            method = item.get("method")

            if col not in X.columns:
                continue

            fitted_item = {"column": col, "method": method}

            if method in ["box-cox", "yeo-johnson"]:
                # Fit PowerTransformer
                try:
                    # Prepare data (Pandas Series/DataFrame)
                    if engine.name == "polars":
                        col_series = X[col].to_pandas()
                        col_df = col_series.to_frame()
                    else:
                        col_series = X[col]
                        col_df = X[[col]]

                    # Box-Cox requires strictly positive
                    if method == "box-cox" and (col_series <= 0).any():
                        logger.warning(
                            f"Skipping Box-Cox for column {col} because it contains non-positive values."
                        )
                        continue

                    # Default to standardize=True for power transforms
                    pt = PowerTransformer(method=method, standardize=True)
                    pt.fit(col_df)

                    fitted_item["lambdas"] = pt.lambdas_.tolist()

                    if hasattr(pt, "_scaler") and pt._scaler:
                        fitted_item["scaler_params"] = {
                            "mean": (
                                pt._scaler.mean_.tolist()
                                if pt._scaler.mean_ is not None
                                else None
                            ),
                            "scale": (
                                pt._scaler.scale_.tolist()
                                if pt._scaler.scale_ is not None
                                else None
                            ),
                        }
                except Exception as e:
                    logger.warning(f"Failed to fit {method} for column {col}: {e}")
                    continue

            fitted_transformations.append(fitted_item)

        return {
            "type": "general_transformation",
            "transformations": fitted_transformations,
        }



from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

setup(
    name="skyulf-core",
    version="0.1.11",
    description="The core machine learning library for Skyulf.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Murat H. Unsal",
    project_urls={
        "Documentation": "https://flyingriverhorse.github.io/Skyulf",
        "Source": "https://github.com/flyingriverhorse/Skyulf",
        "Changelog": "https://github.com/flyingriverhorse/Skyulf/releases",
    },
    include_package_data=True,
    packages=find_packages(),
    install_requires=[
        "pandas>=2.0.0",
        "numpy>=1.24.0",
        "scikit-learn>=1.4.0,<2.0.0",
        "polars>=0.19.0",
        "pydantic>=2.0.0",
        "scipy>=1.10.0",
        "statsmodels>=0.14.0",
    ],
    extras_require={
        "dev": ["pytest", "twine", "build"],
        "viz": ["matplotlib>=3.7.0", "rich>=13.0.0"],
        "eda": [
            "vaderSentiment>=3.3.2",
            "causal-learn>=0.1.3.0",
        ],
        "tuning": [
            "optuna>=3.0.0",
            "optuna-integration>=3.0.0",
        ],
        "preprocessing-imbalanced": ["imbalanced-learn>=0.13.0"],
        "modeling-xgboost": ["xgboost>=2.1.4"],
    },
    python_requires=">=3.9",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
    ],
)



from typing import Any, Dict, Optional, Type
import logging
from .types import NodeMetadataDict

logger = logging.getLogger(__name__)


class NodeRegistry:
    _calculators: Dict[str, Type] = {}
    _appliers: Dict[str, Type] = {}
    _metadata: Dict[str, NodeMetadataDict] = {}

    @classmethod
    def register(
        cls, name: str, applier_cls: Type, metadata: Optional[Dict[str, Any]] = None
    ):
        """
        Decorator to register a Calculator/Applier pair.

        Args:
            name: The unique string identifier for the node (e.g. 'random_forest').
            applier_cls: The class of the Applier (must be passed as we decorate the Calculator).
            metadata: Optional dictionary of UI metadata.
        """

        def wrapper(calculator_cls):
            if name in cls._calculators:
                logger.warning(
                    f"Node '{name}' is being re-registered. Overwriting previous registration."
                )

            cls._calculators[name] = calculator_cls
            cls._appliers[name] = applier_cls

            # 1. Use passed metadata if available
            if metadata:
                cls._metadata[name] = metadata
            # 2. Otherwise check for __node_meta__ (from @node_meta decorator)
            elif hasattr(calculator_cls, "__node_meta__"):
                meta = getattr(calculator_cls, "__node_meta__")
                cls._metadata[name] = {
                    "id": meta.id,
                    "name": meta.name,
                    "category": meta.category,
                    "description": meta.description,
                    "params": meta.params,
                    "tags": meta.tags,
                }

            return calculator_cls

        return wrapper

    @classmethod
    def get_calculator(cls, name: str) -> Type:
        if name not in cls._calculators:
            raise ValueError(
                f"Node '{name}' not found in registry. Available nodes: {list(cls._calculators.keys())}"
            )
        return cls._calculators[name]

    @classmethod
    def get_applier(cls, name: str) -> Type:
        if name not in cls._appliers:
            raise ValueError(f"Node '{name}' not found in registry.")
        return cls._appliers[name]

    @classmethod
    def get_all_metadata(cls) -> Dict[str, Dict[str, Any]]:
        return cls._metadata



"""Regression models."""

from sklearn.ensemble import (
    AdaBoostRegressor,
    GradientBoostingRegressor,
    RandomForestRegressor,
)
from sklearn.linear_model import ElasticNet, Lasso, Ridge, LinearRegression
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR
from sklearn.tree import DecisionTreeRegressor
try:
    from xgboost import XGBRegressor

    XGBOOST_AVAILABLE = True
except ImportError:
    XGBOOST_AVAILABLE = False

from ..core.meta.decorators import node_meta
from ..registry import NodeRegistry
from .sklearn_wrapper import SklearnApplier, SklearnCalculator


# --- Linear Regression ---
class LinearRegressionApplier(SklearnApplier):
    """Linear Regression Applier."""

    pass


@NodeRegistry.register("linear_regression", LinearRegressionApplier)
@node_meta(
    id="linear_regression",
    name="Linear Regression",
    category="Modeling",
    description="Ordinary least squares Linear Regression.",
    params={"fit_intercept": True, "copy_X": True, "n_jobs": -1},
    tags=["requires_scaling"],
)
class LinearRegressionCalculator(SklearnCalculator):
    """Linear Regression Calculator."""

    def __init__(self):
        super().__init__(
            model_class=LinearRegression,
            default_params={
                "fit_intercept": True,
                "copy_X": True,
                "n_jobs": -1,
            },
            problem_type="regression",
        )


# --- Ridge Regression ---
class RidgeRegressionApplier(SklearnApplier):
    """Ridge Regression Applier."""

    pass


@NodeRegistry.register("ridge_regression", RidgeRegressionApplier)
@node_meta(
    id="ridge_regression",
    name="Ridge Regression",
    category="Modeling",
    description="Linear least squares with l2 regularization.",
    params={"alpha": 1.0, "solver": "auto", "random_state": 42},
    tags=["requires_scaling"],
)
class RidgeRegressionCalculator(SklearnCalculator):
    """Ridge Regression Calculator."""

    def __init__(self):
        super().__init__(
            model_class=Ridge,
            default_params={
                "alpha": 1.0,
                "solver": "auto",
                "random_state": 42,
            },
            problem_type="regression",
        )


# --- Random Forest Regressor ---
class RandomForestRegressorApplier(SklearnApplier):
    """Random Forest Regressor Applier."""

    pass


@NodeRegistry.register("random_forest_regressor", RandomForestRegressorApplier)
@node_meta(
    id="random_forest_regressor",
    name="Random Forest Regressor",
    category="Modeling",
    description="Ensemble of decision trees for regression.",
    params={"n_estimators": 50, "max_depth": 10, "min_samples_split": 5}
)
class RandomForestRegressorCalculator(SklearnCalculator):
    """Random Forest Regressor Calculator."""

    def __init__(self):
        super().__init__(
            model_class=RandomForestRegressor,
            default_params={
                "n_estimators": 50,
                "max_depth": 10,
                "min_samples_split": 5,
                "min_samples_leaf": 2,
                "n_jobs": -1,
                "random_state": 42,
            },
            problem_type="regression",
        )


# --- Lasso ---
class LassoRegressionApplier(SklearnApplier):
    """Lasso Regression Applier."""

    pass


@NodeRegistry.register("lasso_regression", LassoRegressionApplier)
@node_meta(
    id="lasso_regression",
    name="Lasso Regression",
    category="Modeling",
    description="Linear Model trained with L1 prior as regularizer.",
    params={"alpha": 1.0, "selection": "cyclic"},
    tags=["requires_scaling"],
)
class LassoRegressionCalculator(SklearnCalculator):
    """Lasso Regression Calculator."""

    def __init__(self):
        super().__init__(
            model_class=Lasso,
            default_params={"alpha": 1.0, "selection": "cyclic", "random_state": 42},
            problem_type="regression",
        )


# --- ElasticNet ---
class ElasticNetRegressionApplier(SklearnApplier):
    """ElasticNet Regression Applier."""

    pass


@NodeRegistry.register("elasticnet_regression", ElasticNetRegressionApplier)
@node_meta(
    id="elasticnet_regression",
    name="ElasticNet Regression",
    category="Modeling",
    description="Linear regression with combined L1 and L2 priors.",
    params={"alpha": 1.0, "l1_ratio": 0.5, "selection": "cyclic"},
    tags=["requires_scaling"],
)
class ElasticNetRegressionCalculator(SklearnCalculator):
    """ElasticNet Regression Calculator."""

    def __init__(self):
        super().__init__(
            model_class=ElasticNet,
            default_params={
                "alpha": 1.0,
                "l1_ratio": 0.5,
                "selection": "cyclic",
                "random_state": 42,
            },
            problem_type="regression",
        )


# --- SVR ---
class SVRApplier(SklearnApplier):
    """SVR Applier."""

    pass


@NodeRegistry.register("svr", SVRApplier)
@node_meta(
    id="svr",
    name="Support Vector Regressor",
    category="Modeling",
    description="Epsilon-Support Vector Regression.",
    params={"C": 1.0, "kernel": "rbf", "gamma": "scale"},
    tags=["requires_scaling"],
)
class SVRCalculator(SklearnCalculator):
    """SVR Calculator."""

    def __init__(self):
        super().__init__(
            model_class=SVR,
            default_params={"C": 1.0, "kernel": "rbf", "gamma": "scale"},
            problem_type="regression",
        )


# --- K-Neighbors ---
class KNeighborsRegressorApplier(SklearnApplier):
    """K-Neighbors Regressor Applier."""

    pass


@NodeRegistry.register("k_neighbors_regressor", KNeighborsRegressorApplier)
@node_meta(
    id="k_neighbors_regressor",
    name="K-Neighbors Regressor",
    category="Modeling",
    description="Regression based on k-nearest neighbors.",
    params={"n_neighbors": 5, "weights": "uniform", "algorithm": "auto"},
    tags=["requires_scaling"],
)
class KNeighborsRegressorCalculator(SklearnCalculator):
    """K-Neighbors Regressor Calculator."""

    def __init__(self):
        super().__init__(
            model_class=KNeighborsRegressor,
            default_params={
                "n_neighbors": 5,
                "weights": "uniform",
                "algorithm": "auto",
                "n_jobs": -1,
            },
            problem_type="regression",
        )


# --- Decision Tree ---
class DecisionTreeRegressorApplier(SklearnApplier):
    """Decision Tree Regressor Applier."""

    pass


@NodeRegistry.register("decision_tree_regressor", DecisionTreeRegressorApplier)
@node_meta(
    id="decision_tree_regressor",
    name="Decision Tree Regressor",
    category="Modeling",
    description="A decision tree regressor.",
    params={"max_depth": None, "min_samples_split": 2, "criterion": "squared_error"},
)
class DecisionTreeRegressorCalculator(SklearnCalculator):
    """Decision Tree Regressor Calculator."""

    def __init__(self):
        super().__init__(
            model_class=DecisionTreeRegressor,
            default_params={
                "max_depth": None,
                "min_samples_split": 2,
                "criterion": "squared_error",
                "random_state": 42,
            },
            problem_type="regression",
        )


# --- Gradient Boosting ---
class GradientBoostingRegressorApplier(SklearnApplier):
    """Gradient Boosting Regressor Applier."""

    pass


@NodeRegistry.register("gradient_boosting_regressor", GradientBoostingRegressorApplier)
@node_meta(
    id="gradient_boosting_regressor",
    name="Gradient Boosting Regressor",
    category="Modeling",
    description="Gradient Boosting for regression.",
    params={"n_estimators": 100, "learning_rate": 0.1, "max_depth": 3},
)
class GradientBoostingRegressorCalculator(SklearnCalculator):
    """Gradient Boosting Regressor Calculator."""

    def __init__(self):
        super().__init__(
            model_class=GradientBoostingRegressor,
            default_params={
                "n_estimators": 100,
                "learning_rate": 0.1,
                "max_depth": 3,
                "random_state": 42,
            },
            problem_type="regression",
        )


# --- AdaBoost ---
class AdaBoostRegressorApplier(SklearnApplier):
    """AdaBoost Regressor Applier."""

    pass


@NodeRegistry.register("adaboost_regressor", AdaBoostRegressorApplier)
@node_meta(
    id="adaboost_regressor",
    name="AdaBoost Regressor",
    category="Modeling",
    description="An AdaBoost regressor.",
    params={"n_estimators": 50, "learning_rate": 1.0},
)
class AdaBoostRegressorCalculator(SklearnCalculator):
    """AdaBoost Regressor Calculator."""

    def __init__(self):
        super().__init__(
            model_class=AdaBoostRegressor,
            default_params={
                "n_estimators": 50,
                "learning_rate": 1.0,
                "random_state": 42,
            },
            problem_type="regression",
        )


# --- XGBoost ---
if XGBOOST_AVAILABLE:
    class XGBRegressorApplier(SklearnApplier):
        """XGBoost Regressor Applier."""

        pass


    @NodeRegistry.register("xgboost_regressor", XGBRegressorApplier)
    @node_meta(
        id="xgboost_regressor",
        name="XGBoost Regressor",
        category="Modeling",
        description="Extreme Gradient Boosting regressor.",
        params={"n_estimators": 100, "max_depth": 6, "learning_rate": 0.3},
    )
    class XGBRegressorCalculator(SklearnCalculator):
        """XGBoost Regressor Calculator."""

        def __init__(self):
            super().__init__(
                model_class=XGBRegressor,
                default_params={
                    "n_estimators": 100,
                    "max_depth": 6,
                    "learning_rate": 0.3,
                    "n_jobs": -1,
                    "random_state": 42,
                    "eval_metric": "rmse",
                },
                problem_type="regression",
            )



from typing import Any, Dict, Tuple, Union

import pandas as pd

from ..registry import NodeRegistry
from ..core.meta.decorators import node_meta
from ..utils import detect_numeric_columns, unpack_pipeline_input
from .base import BaseApplier, BaseCalculator
from ..engines import SkyulfDataFrame, get_engine


class DatasetProfileApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        # Inspection nodes do not modify data
        return df


@NodeRegistry.register("DatasetProfile", DatasetProfileApplier)
@node_meta(
    id="DatasetProfile",
    name="Dataset Profile",
    category="Inspection",
    description="Generate a statistical profile of the dataset.",
    params={}
)
class DatasetProfileCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        # Generate a lightweight dataset profile
        X, _, _ = unpack_pipeline_input(df)
        engine = get_engine(X)

        profile: Dict[str, Any] = {}

        if engine.name == "polars":
            import polars as pl

            profile["rows"] = len(X)
            profile["columns"] = len(X.columns)
            profile["dtypes"] = {
                col: str(dtype) for col, dtype in zip(X.columns, X.dtypes)
            }
            profile["missing"] = {col: X[col].null_count() for col in X.columns}

            # Numeric stats
            numeric_cols = [
                col
                for col, dtype in zip(X.columns, X.dtypes)
                if dtype in [pl.Float64, pl.Float32, pl.Int64, pl.Int32]
            ]
            if numeric_cols:
                # Polars describe returns a DataFrame
                desc_df = X.select(numeric_cols).describe()
                # Convert to list of dicts
                desc_dicts = desc_df.to_dicts()
                
                stats = {}
                # Initialize stats dicts
                for col in numeric_cols:
                    stats[col] = {}
                
                for row in desc_dicts:
                    # Polars < 0.19 uses "describe", newer uses "statistic"
                    metric = row.get("describe") or row.get("statistic")
                    if not metric:
                        continue
                        
                    for col in numeric_cols:
                        if col in row:
                            stats[col][metric] = row[col]
                            
                profile["numeric_stats"] = stats
        else:
            # Pandas logic
            profile["rows"] = len(X)
            profile["columns"] = len(X.columns)
            profile["dtypes"] = X.dtypes.astype(str).to_dict()
            profile["missing"] = X.isna().sum().to_dict()

            numeric_cols = detect_numeric_columns(X)
            if numeric_cols:
                desc = X[numeric_cols].describe().to_dict()
                profile["numeric_stats"] = desc

        return {"type": "dataset_profile", "profile": profile}


class DataSnapshotApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        return df


@NodeRegistry.register("DataSnapshot", DataSnapshotApplier)
@node_meta(
    id="DataSnapshot",
    name="Data Snapshot",
    category="Inspection",
    description="Take a snapshot of the first N rows of the dataset.",
    params={"n_rows": 5}
)
class DataSnapshotCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        X, _, _ = unpack_pipeline_input(df)
        engine = get_engine(X)

        n = config.get("n_rows", 5)
        
        if engine.name == "polars":
            snapshot = X.head(n).to_dicts()
        else:
            snapshot = X.head(n).to_dict(orient="records")

        return {"type": "data_snapshot", "snapshot": snapshot}



"""
Skyulf Core SDK
"""

from .data.dataset import SplitDataset
from .pipeline import SkyulfPipeline
from .preprocessing.pipeline import FeatureEngineer

__version__ = "0.1.11"

__all__ = [
    "SkyulfPipeline",
    "SplitDataset",
    "FeatureEngineer",
]



import builtins
import math
from difflib import SequenceMatcher
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from sklearn.preprocessing import PolynomialFeatures

from ..registry import NodeRegistry
from ..core.meta.decorators import node_meta
from ..utils import detect_numeric_columns, pack_pipeline_output, unpack_pipeline_input
from .base import BaseApplier, BaseCalculator
from ..engines import SkyulfDataFrame, get_engine

# --- Optional Dependencies ---
try:
    from rapidfuzz import fuzz  # type: ignore

    _HAS_RAPIDFUZZ = True
except ImportError:
    fuzz = None
    _HAS_RAPIDFUZZ = False

# --- Constants ---
DEFAULT_EPSILON = 1e-9
FEATURE_MATH_ALLOWED_TYPES = {
    "arithmetic",
    "ratio",
    "similarity",
    "datetime_extract",
    "group_agg",
    "polynomial",
}
ALLOWED_DATETIME_FEATURES = {
    "year",
    "quarter",
    "month",
    "month_name",
    "week",
    "day",
    "day_name",
    "weekday",
    "is_weekend",
    "hour",
    "minute",
    "second",
    "season",
    "time_of_day",
}

# --- Helpers ---


def _coerce_float(value: Any) -> Optional[float]:
    if value is None:
        return None
    try:
        val = float(value)
        return val if not math.isnan(val) else None
    except (TypeError, ValueError):
        return None


def _safe_divide(
    numerator: pd.Series, denominator: pd.Series, epsilon: float
) -> pd.Series:
    adjusted = denominator.copy()
    adjusted = adjusted.replace({0: epsilon, -0.0: epsilon})
    adjusted = adjusted.fillna(epsilon)
    # Avoid division by zero or near-zero
    mask = adjusted.abs() < epsilon
    if mask.any():
        adjusted[mask] = epsilon
    return numerator / adjusted


def _compute_similarity_score(a: Any, b: Any, method: str) -> float:
    text_a = "" if pd.isna(a) else str(a)
    text_b = "" if pd.isna(b) else str(b)
    if not text_a and not text_b:
        return 100.0
    if not text_a or not text_b:
        return 0.0

    if _HAS_RAPIDFUZZ:
        if method == "token_sort_ratio":
            return float(fuzz.token_sort_ratio(text_a, text_b))
        if method == "token_set_ratio":
            return float(fuzz.token_set_ratio(text_a, text_b))
        return float(fuzz.ratio(text_a, text_b))

    # Fallback
    return SequenceMatcher(None, text_a, text_b).ratio() * 100.0


# --- Polynomial Features ---


class PolynomialFeaturesApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        cols = params.get("columns", [])
        valid_cols = [c for c in cols if c in X.columns]

        if not valid_cols:
            return pack_pipeline_output(X, y, is_tuple)

        degree = params.get("degree", 2)
        interaction_only = params.get("interaction_only", False)
        include_bias = params.get("include_bias", False)
        include_input_features = params.get("include_input_features", False)
        output_prefix = params.get("output_prefix", "poly")

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            X_pl: Any = X
            
            # For PolynomialFeatures, we use the pandas/sklearn implementation via conversion
            # to ensure full compatibility with complex degree/interaction logic.
            X_pd = X_pl.to_pandas()
            
            poly = PolynomialFeatures(
                degree=degree, interaction_only=interaction_only, include_bias=include_bias
            )
            poly.fit(X_pd[valid_cols])
            
            transformed = poly.transform(X_pd[valid_cols])
            feature_names = poly.get_feature_names_out(valid_cols)
            
            # Filter logic
            indices_to_keep = []
            for i, p in enumerate(poly.powers_):
                deg = sum(p)
                if deg == 1 and not include_input_features:
                    continue
                indices_to_keep.append(i)
                
            if not indices_to_keep:
                return pack_pipeline_output(X, y, is_tuple)
                
            transformed = transformed[:, indices_to_keep]
            feature_names = feature_names[indices_to_keep]
            
            new_names = []
            for name in feature_names:
                clean_name = name.replace(" ", "_").replace("^", "_pow_")
                new_names.append(f"{output_prefix}_{clean_name}")
            
            # Create Polars DataFrame
            df_poly = pl.DataFrame(transformed, schema=new_names)
            
            # Horizontal concat
            X_out = pl.concat([X_pl, df_poly], how="horizontal")
            
            return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        poly = PolynomialFeatures(
            degree=degree, interaction_only=interaction_only, include_bias=include_bias
        )
        poly.fit(X[valid_cols])  # Cheap fit

        transformed = poly.transform(X[valid_cols])

        # Handle case where sklearn is configured to output pandas DataFrames
        if hasattr(transformed, "iloc"):
            transformed = transformed.values

        feature_names = poly.get_feature_names_out(valid_cols)

        # Filter out original features (degree 1) if not requested
        # We use powers_ to determine the degree of each output feature
        indices_to_keep = []
        for i, p in enumerate(poly.powers_):
            deg = sum(p)
            # If degree is 1 (linear term) and we don't want to include input features, skip it
            if deg == 1 and not include_input_features:
                continue
            indices_to_keep.append(i)

        if not indices_to_keep:
            return pack_pipeline_output(X, y, is_tuple)

        transformed = transformed[:, indices_to_keep]
        feature_names = feature_names[indices_to_keep]

        # Rename features to be more readable and avoid collisions
        # We convert sklearn's "col1 col2" format to "prefix_col1_col2"

        new_names = []
        for name in feature_names:
            # Clean up name
            clean_name = name.replace(" ", "_").replace("^", "_pow_")
            new_names.append(f"{output_prefix}_{clean_name}")

        df_poly = pd.DataFrame(transformed, columns=new_names, index=X.index)

        df_out = X.copy()
        # We no longer need to check include_input_features here as we filtered them above

        # Concatenate
        df_out = pd.concat([df_out, df_poly], axis=1)

        return pack_pipeline_output(df_out, y, is_tuple)


@NodeRegistry.register("PolynomialFeatures", PolynomialFeaturesApplier)
@NodeRegistry.register("PolynomialFeaturesNode", PolynomialFeaturesApplier)
@node_meta(
    id="PolynomialFeatures",
    name="Polynomial Features",
    category="Feature Engineering",
    description="Generate polynomial and interaction features.",
    params={"degree": 2, "interaction_only": False, "include_bias": False}
)
class PolynomialFeaturesCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        # Extract configuration parameters
        # We support standard polynomial features settings like degree and interaction_only

        X, _, _ = unpack_pipeline_input(df)
        engine = get_engine(X)

        cols = config.get("columns", [])
        auto_detect = config.get("auto_detect", False)

        if not cols and auto_detect:
            cols = detect_numeric_columns(X)

        cols = [c for c in cols if c in X.columns]

        if not cols:
            return {}

        degree = config.get("degree", 2)
        interaction_only = config.get("interaction_only", False)
        include_bias = config.get("include_bias", False)

        # We use sklearn to get feature names
        # Ensure X is compatible with sklearn (Pandas/Numpy)
        if engine.name == "polars":
            X_pl: Any = X
            X_fit = X_pl.select(cols).to_pandas()
        else:
            X_fit = X[cols]

        poly = PolynomialFeatures(
            degree=degree, interaction_only=interaction_only, include_bias=include_bias
        )
        poly.fit(X_fit)

        return {
            "type": "polynomial_features",
            "columns": cols,
            "degree": degree,
            "interaction_only": interaction_only,
            "include_bias": include_bias,
            "include_input_features": config.get("include_input_features", False),
            "output_prefix": config.get("output_prefix", "poly"),
            "feature_names": poly.get_feature_names_out(cols).tolist(),
        }


# --- Feature Generation ---


class FeatureGenerationApplier(BaseApplier):
    def apply(  # noqa: C901
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        operations = params.get("operations", [])
        epsilon = params.get("epsilon", DEFAULT_EPSILON)
        allow_overwrite = params.get("allow_overwrite", False)

        if not operations:
            return pack_pipeline_output(X, y, is_tuple)

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            X_pl: Any = X
            
            X_out = X_pl
            
            for i, op in enumerate(operations):
                op_type = op.get("operation_type", "arithmetic")
                method = op.get("method")
                input_cols = op.get("input_columns", [])
                secondary_cols = op.get("secondary_columns", [])
                constants = op.get("constants", [])
                output_col = op.get("output_column")
                output_prefix = op.get("output_prefix")
                fillna = op.get("fillna")
                round_digits = op.get("round_digits")

                # Resolve output name
                if not output_col:
                    base = f"{op_type}_{i}"
                    if output_prefix:
                        base = f"{output_prefix}_{base}"
                    output_col = base

                if output_col in X_out.columns and not allow_overwrite:
                    j = 1
                    while f"{output_col}_{j}" in X_out.columns:
                        j += 1
                    output_col = f"{output_col}_{j}"

                try:
                    expr = None
                    
                    if op_type == "arithmetic":
                        valid_inputs = [c for c in input_cols if c in X_out.columns]
                        valid_secondary = [c for c in secondary_cols if c in X_out.columns]
                        all_cols = valid_inputs + valid_secondary
                        
                        if not all_cols and not constants:
                            continue
                            
                        fill_val = fillna if fillna is not None else 0
                        col_exprs = [pl.col(c).cast(pl.Float64).fill_null(fill_val) for c in all_cols]
                        const_vals = [float(c) for c in constants]
                        
                        if method == "add":
                            expr = pl.sum_horizontal(col_exprs) + sum(const_vals)
                            
                        elif method == "subtract":
                            if col_exprs:
                                expr = col_exprs[0]
                                others = col_exprs[1:]
                            else:
                                expr = pl.lit(0.0)
                                others = []
                                
                            for e in others:
                                expr = expr - e
                            for c in const_vals:
                                expr = expr - c
                                
                        elif method == "multiply":
                            expr = pl.lit(1.0)
                            for e in col_exprs:
                                expr = expr * e
                            for c in const_vals:
                                expr = expr * c
                                
                        elif method == "divide":
                            if col_exprs:
                                expr = col_exprs[0]
                                others = col_exprs[1:]
                            elif const_vals:
                                expr = pl.lit(const_vals[0])
                                others = []
                                const_vals = const_vals[1:]
                            else:
                                continue
                                
                            def safe_denom(d):
                                return pl.when(d.abs() < epsilon).then(epsilon).otherwise(d)
                                
                            for e in others:
                                expr = expr / safe_denom(e)
                            for c in const_vals:
                                c_val = c if abs(c) > epsilon else epsilon
                                expr = expr / c_val

                    elif op_type == "ratio":
                        nums = [pl.col(c).cast(pl.Float64).fill_null(0) for c in input_cols if c in X_out.columns]
                        dens = [pl.col(c).cast(pl.Float64).fill_null(0) for c in secondary_cols if c in X_out.columns]
                        
                        if not nums or not dens:
                            continue
                            
                        num_sum = pl.sum_horizontal(nums)
                        den_sum = pl.sum_horizontal(dens)
                        
                        expr = num_sum / pl.when(den_sum.abs() < epsilon).then(epsilon).otherwise(den_sum)

                    elif op_type == "similarity":
                        col_a = input_cols[0] if input_cols else None
                        col_b = secondary_cols[0] if secondary_cols else (input_cols[1] if len(input_cols) > 1 else None)
                        
                        if col_a and col_b and col_a in X_out.columns and col_b in X_out.columns:
                            def sim_func(struct_val):
                                a = struct_val.get("a")
                                b = struct_val.get("b")
                                return _compute_similarity_score(a, b, method)
                                
                            expr = pl.struct([pl.col(col_a).alias("a"), pl.col(col_b).alias("b")]).map_elements(sim_func, return_dtype=pl.Float64)

                    elif op_type == "datetime_extract":
                        valid_inputs = [c for c in input_cols if c in X_out.columns]
                        features = op.get("datetime_features", [])
                        
                        dt_exprs = []
                        for col in valid_inputs:
                            dtype = X_out.schema[col]
                            base_dt = pl.col(col)
                            if dtype == pl.String:
                                base_dt = pl.col(col).str.to_datetime(strict=False)
                            
                            for feat in features:
                                feat_name = f"{col}_{feat}"
                                val = None
                                if feat == "year": val = base_dt.dt.year()
                                elif feat == "month": val = base_dt.dt.month()
                                elif feat == "day": val = base_dt.dt.day()
                                elif feat == "hour": val = base_dt.dt.hour()
                                elif feat == "minute": val = base_dt.dt.minute()
                                elif feat == "second": val = base_dt.dt.second()
                                elif feat == "quarter": val = base_dt.dt.quarter()
                                elif feat == "dayofweek" or feat == "weekday": val = base_dt.dt.weekday() - 1
                                elif feat == "is_weekend": val = (base_dt.dt.weekday() >= 6).cast(pl.Int64)
                                
                                if val is not None:
                                    dt_exprs.append(val.alias(feat_name))
                        
                        if dt_exprs:
                            X_out = X_out.with_columns(dt_exprs)
                        continue

                    if expr is not None:
                        if round_digits is not None:
                            expr = expr.round(round_digits)
                        X_out = X_out.with_columns(expr.alias(output_col))

                except Exception:
                    pass
            
            return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        df_out = X.copy()

        for i, op in enumerate(operations):
            op_type = op.get("operation_type", "arithmetic")
            method = op.get("method")
            input_cols = op.get("input_columns", [])
            secondary_cols = op.get("secondary_columns", [])
            constants = op.get("constants", [])
            output_col = op.get("output_column")
            output_prefix = op.get("output_prefix")
            fillna = op.get("fillna")
            round_digits = op.get("round_digits")

            # Resolve output name
            if not output_col:
                # Generate fallback
                base = f"{op_type}_{i}"
                if output_prefix:
                    base = f"{output_prefix}_{base}"
                output_col = base

            if output_col in df_out.columns and not allow_overwrite:
                # Avoid overwriting existing columns by appending a numeric suffix
                j = 1
                while f"{output_col}_{j}" in df_out.columns:
                    j += 1
                output_col = f"{output_col}_{j}"

            try:
                result = None

                if op_type == "arithmetic":
                    # Ensure inputs exist
                    valid_inputs = [c for c in input_cols if c in df_out.columns]
                    valid_secondary = [c for c in secondary_cols if c in df_out.columns]
                    all_cols = valid_inputs + valid_secondary
                    if not all_cols and not constants:
                        continue

                    series_list = [
                        pd.to_numeric(df_out[c], errors="coerce").fillna(
                            fillna if fillna is not None else 0
                        )
                        for c in all_cols
                    ]

                    if method == "add":
                        res: pd.Series = pd.Series(0.0, index=df_out.index)
                        for s in series_list:
                            res = res.add(s, fill_value=0)
                        for c in constants:
                            val = builtins.float(c)
                            res = res.add(val)
                        result = res
                    elif method == "subtract":
                        res_sub: pd.Series
                        if series_list:
                            res_sub = series_list[0].copy()
                            others = series_list[1:]
                        else:
                            res_sub = pd.Series(0.0, index=df_out.index)
                            others = []
                        for s in others:
                            res_sub = res_sub.subtract(s, fill_value=0)
                        for c in constants:
                            val = builtins.float(c)
                            res_sub = res_sub.sub(val)
                        result = res_sub
                    elif method == "multiply":
                        res_mul: pd.Series = pd.Series(1.0, index=df_out.index)
                        for s in series_list:
                            res_mul = res_mul.multiply(s, fill_value=1)
                        for c in constants:
                            val = builtins.float(c)
                            res_mul = res_mul.mul(val)
                        result = res_mul
                    elif method == "divide":
                        if series_list:
                            res = series_list[0].copy()
                            others = series_list[1:]
                        elif constants:
                            res = pd.Series(constants[0], index=df_out.index)
                            others = []
                            constants = constants[1:]
                        else:
                            continue

                        for s in others:
                            res = _safe_divide(res, s, epsilon)
                        for c in constants:
                            c_val = builtins.float(c)
                            # For division, we need to handle epsilon check
                            # res = res / (c_val if abs(c_val) > epsilon else epsilon)
                            denom = c_val if abs(c_val) > epsilon else epsilon
                            res = res.div(denom)
                        result = res

                elif op_type == "ratio":
                    # input_cols (numerator) / secondary_cols (denominator)
                    # Sum of numerators / Sum of denominators
                    nums = [
                        pd.to_numeric(df_out[c], errors="coerce").fillna(0)
                        for c in input_cols
                        if c in df_out.columns
                    ]
                    dens = [
                        pd.to_numeric(df_out[c], errors="coerce").fillna(0)
                        for c in secondary_cols
                        if c in df_out.columns
                    ]

                    if not nums or not dens:
                        continue

                    num_sum = pd.Series(0.0, index=df_out.index)
                    for s in nums:
                        num_sum = num_sum.add(s, fill_value=0)

                    den_sum = pd.Series(0.0, index=df_out.index)
                    for s in dens:
                        den_sum = den_sum.add(s, fill_value=0)

                    result = _safe_divide(num_sum, den_sum, epsilon)

                elif op_type == "similarity":
                    # input_cols[0] vs secondary_cols[0] (or input_cols[1])
                    col_a = input_cols[0] if input_cols else None
                    col_b = (
                        secondary_cols[0]
                        if secondary_cols
                        else (input_cols[1] if len(input_cols) > 1 else None)
                    )

                    if (
                        not col_a
                        or not col_b
                        or col_a not in df_out.columns
                        or col_b not in df_out.columns
                    ):
                        continue

                    result = df_out.apply(
                        lambda row: _compute_similarity_score(
                            row[col_a], row[col_b], method
                        ),
                        axis=1,
                    )

                elif op_type == "datetime_extract":
                    valid_inputs = [c for c in input_cols if c in df_out.columns]
                    features = op.get("datetime_features", [])

                    for col in valid_inputs:
                        try:
                            dt = pd.to_datetime(df_out[col], errors="coerce")
                            for feat in features:
                                feat_name = f"{col}_{feat}"
                                if feat == "year":
                                    val = dt.dt.year
                                elif feat == "month":
                                    val = dt.dt.month
                                elif feat == "day":
                                    val = dt.dt.day
                                elif feat == "hour":
                                    val = dt.dt.hour
                                elif feat == "minute":
                                    val = dt.dt.minute
                                elif feat == "second":
                                    val = dt.dt.second
                                elif feat == "quarter":
                                    val = dt.dt.quarter
                                elif feat == "dayofweek" or feat == "weekday":
                                    val = dt.dt.dayofweek
                                elif feat == "is_weekend":
                                    val = (dt.dt.dayofweek >= 5).astype(int)
                                else:
                                    continue

                                df_out[feat_name] = val
                        except Exception:
                            pass
                    # datetime_extract usually generates multiple columns, so we might not set "result"
                    # unless we want to return one specific thing. V1 generates multiple.
                    continue

                if result is not None:
                    if round_digits is not None:
                        result = result.round(round_digits)
                    df_out[output_col] = result

            except Exception:
                pass

        return pack_pipeline_output(df_out, y, is_tuple)


@NodeRegistry.register("FeatureGeneration", FeatureGenerationApplier)
@NodeRegistry.register("FeatureMath", FeatureGenerationApplier)
@NodeRegistry.register("FeatureGenerationNode", FeatureGenerationApplier)
@node_meta(
    id="FeatureGenerationNode",
    name="Feature Generation (Math)",
    category="Feature Engineering",
    description="Generate new features using mathematical operations.",
    params={"operations": []}
)
class FeatureGenerationCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        # Config:
        # operations: List[Dict]
        # epsilon: float
        # allow_overwrite: bool

        return {
            "type": "feature_generation",
            "operations": config.get("operations", []),
            "epsilon": config.get("epsilon", DEFAULT_EPSILON),
            "allow_overwrite": config.get("allow_overwrite", False),
        }



"""Tuning configuration schemas."""

from dataclasses import dataclass, field
from typing import Any, Dict, List, Literal, Optional


@dataclass
class TuningConfig:
    """Configuration for hyperparameter tuning."""

    strategy: Literal["grid", "random", "optuna", "halving_grid", "halving_random"] = (
        "random"
    )
    metric: str = "accuracy"  # or 'mse', 'f1', etc.
    n_trials: int = 10
    timeout: Optional[int] = None  # Seconds
    search_space: Dict[str, List[Any]] = field(
        default_factory=dict
    )  # e.g. {"C": [0.1, 1.0, 10.0]}
    cv_enabled: bool = True
    cv_folds: int = 5
    cv_type: Literal[
        "k_fold", "stratified_k_fold", "time_series_split", "shuffle_split"
    ] = "k_fold"
    cv_shuffle: bool = True
    cv_random_state: int = 42
    random_state: int = 42


@dataclass
class TuningResult:
    """Result of a tuning session."""

    best_params: Dict[str, Any]
    best_score: float
    n_trials: int
    trials: List[Dict[str, Any]]  # List of {params, score}



from typing import Any, Dict, Optional
from dataclasses import dataclass, field

@dataclass
class NodeMetadata:
    id: str
    name: str
    category: str
    description: str
    params: Dict[str, Any] = field(default_factory=dict)
    tags: list[str] = field(default_factory=list)


def node_meta(
    id: str,
    name: str,
    category: str,
    description: str,
    params: Optional[Dict[str, Any]] = None,
    tags: Optional[list[str]] = None,
):
    def decorator(cls):
        cls.__node_meta__ = NodeMetadata(
            id=id,
            name=name,
            category=category,
            description=description,
            params=params or {},
            tags=tags or [],
        )
        return cls

    return decorator



import polars as pl
import numpy as np
from typing import Dict, Any, Optional, List, Union
from pydantic import BaseModel

try:
    from scipy.stats import ks_2samp, wasserstein_distance, entropy
    SCIPY_AVAILABLE = True
except ImportError:
    SCIPY_AVAILABLE = False

class DriftMetric(BaseModel):
    metric: str
    value: float
    has_drift: bool
    threshold: float

class DriftBin(BaseModel):
    bin_start: float
    bin_end: float
    reference_count: int
    current_count: int

class DriftDistribution(BaseModel):
    bins: List[DriftBin]

class ColumnDrift(BaseModel):
    column: str
    metrics: List[DriftMetric]
    drift_detected: bool
    suggestions: List[str] = []
    distribution: Optional[DriftDistribution] = None

class DriftReport(BaseModel):
    reference_rows: int
    current_rows: int
    drifted_columns_count: int
    column_drifts: Dict[str, ColumnDrift]
    missing_columns: List[str] = []
    new_columns: List[str] = []

class DriftCalculator:
    """
    Calculates data drift between a reference dataset (training) and current dataset (production).
    Uses Polars for efficient data processing.
    """
    
    def __init__(self, reference_df: pl.DataFrame, current_df: pl.DataFrame):
        self.reference_df = reference_df
        self.current_df = current_df
        self.common_columns = [
            col for col in reference_df.columns 
            if col in current_df.columns
        ]

    def calculate_drift(self, thresholds: Optional[Dict[str, float]] = None) -> DriftReport:
        """
        Calculates drift for all common columns.
        """
        if not SCIPY_AVAILABLE:
            raise ImportError("scipy is required for drift calculation")

        default_thresholds = {
            "psi": 0.2,
            "ks": 0.05, # p-value < 0.05 means distributions are different
            "wasserstein": 0.1, # Heuristic, depends on scale
            "kl_divergence": 0.1
        }
        thresholds = {**default_thresholds, **(thresholds or {})}
        
        column_drifts = {}
        drifted_count = 0
        
        # Schema Drift Detection
        ref_cols = set(self.reference_df.columns)
        curr_cols = set(self.current_df.columns)
        missing_columns = list(ref_cols - curr_cols)
        new_columns = list(curr_cols - ref_cols)
        
        for col in self.common_columns:
            # Skip non-numeric for now (except PSI which can handle categorical with binning)
            # For simplicity, let's focus on numeric columns first
            dtype = self.reference_df[col].dtype
            if dtype not in [pl.Float32, pl.Float64, pl.Int32, pl.Int64]:
                continue
            
            # Ensure current data is also numeric or castable to the reference type
            curr_series = self.current_df[col]
            if curr_series.dtype != dtype:
                try:
                    # Try to cast current to match reference (e.g. Int to Float, or String to Float)
                    curr_series = curr_series.cast(dtype, strict=False)
                except Exception:
                    # If casting fails completely (unlikely with strict=False), skip
                    continue
                
            ref_data = self.reference_df[col].drop_nulls().to_numpy()
            curr_data = curr_series.drop_nulls().to_numpy()
            
            if len(ref_data) == 0 or len(curr_data) == 0:
                continue
                
            metrics = []
            is_drifted = False
            
            # 1. Wasserstein Distance
            wd = wasserstein_distance(ref_data, curr_data)
            # Normalize WD? It's scale dependent. 
            # Simple normalization: divide by std of reference
            std_ref = np.std(ref_data)
            norm_wd = wd / std_ref if std_ref > 0 else wd
            
            wd_drift = norm_wd > thresholds["wasserstein"]
            metrics.append(DriftMetric(
                metric="wasserstein_distance",
                value=float(wd),
                has_drift=wd_drift,
                threshold=thresholds["wasserstein"]
            ))
            if wd_drift: is_drifted = True
            
            # 2. KS Test
            ks_stat, ks_p = ks_2samp(ref_data, curr_data)
            ks_drift = ks_p < thresholds["ks"]
            metrics.append(DriftMetric(
                metric="ks_test_p_value",
                value=float(ks_p),
                has_drift=ks_drift,
                threshold=thresholds["ks"]
            ))
            if ks_drift: is_drifted = True
            
            # 3. PSI (Population Stability Index)
            psi_val = self._calculate_psi(ref_data, curr_data)
            psi_drift = psi_val > thresholds["psi"]
            metrics.append(DriftMetric(
                metric="psi",
                value=float(psi_val),
                has_drift=psi_drift,
                threshold=thresholds["psi"]
            ))
            if psi_drift: is_drifted = True

            # 4. KL Divergence
            kl_val = self._calculate_kl(ref_data, curr_data)
            kl_drift = kl_val > thresholds["kl_divergence"]
            metrics.append(DriftMetric(
                metric="kl_divergence",
                value=float(kl_val),
                has_drift=kl_drift,
                threshold=thresholds["kl_divergence"]
            ))
            if kl_drift: is_drifted = True
            
            # Generate Suggestions
            suggestions = []
            if is_drifted:
                if psi_val > 0.25:
                    suggestions.append("Critical population shift detected (PSI > 0.25). Immediate model retraining is recommended.")
                elif psi_val > 0.1:
                    suggestions.append("Moderate population shift detected. Monitor model performance closely.")
                
                if ks_drift and not psi_drift:
                    suggestions.append("Statistical distribution has changed, but population stability is acceptable. Check for outliers.")
                
                if wd_drift:
                    suggestions.append("Significant change in feature scale or shape detected. Verify data preprocessing steps.")

            # Calculate Distribution (Histogram)
            distribution = self._calculate_distribution(ref_data, curr_data)

            column_drifts[col] = ColumnDrift(
                column=col,
                metrics=metrics,
                drift_detected=is_drifted,
                suggestions=suggestions,
                distribution=distribution
            )
            
            if is_drifted:
                drifted_count += 1
                
        return DriftReport(
            reference_rows=len(self.reference_df),
            current_rows=len(self.current_df),
            drifted_columns_count=drifted_count,
            column_drifts=column_drifts,
            missing_columns=missing_columns,
            new_columns=new_columns
        )

    def _calculate_distribution(self, ref_data: np.ndarray, curr_data: np.ndarray, bins: int = 20) -> DriftDistribution:
        """
        Calculates histogram bins for reference and current data using the same range.
        """
        try:
            # Determine global min/max
            min_val = min(np.min(ref_data), np.min(curr_data))
            max_val = max(np.max(ref_data), np.max(curr_data))
            
            # Handle constant case
            if min_val == max_val:
                min_val -= 0.5
                max_val += 0.5

            # Compute histogram for both using the same range
            ref_hist, bin_edges = np.histogram(ref_data, bins=bins, range=(min_val, max_val))
            curr_hist, _ = np.histogram(curr_data, bins=bins, range=(min_val, max_val))
            
            drift_bins = []
            for i in range(len(ref_hist)):
                drift_bins.append(DriftBin(
                    bin_start=float(bin_edges[i]),
                    bin_end=float(bin_edges[i+1]),
                    reference_count=int(ref_hist[i]),
                    current_count=int(curr_hist[i])
                ))
                
            return DriftDistribution(bins=drift_bins)
        except Exception:
            return DriftDistribution(bins=[])

    def _calculate_psi(self, expected: np.ndarray, actual: np.ndarray, buckets: int = 10) -> float:
        """
        Calculate Population Stability Index (PSI).
        """
        def scale_range(input, min, max):
            input += -(np.min(input))
            input /= np.max(input) / (max - min)
            input += min
            return input

        breakpoints = np.arange(0, buckets + 1) / (buckets) * 100
        
        if len(expected) == 0 or len(actual) == 0:
            return 0.0

        # Use percentiles from expected (reference) to define bins
        try:
            # Handle constant arrays
            if np.min(expected) == np.max(expected):
                return 0.0
                
            breakpoints = np.percentile(expected, breakpoints)
            
            # Ensure unique breakpoints
            breakpoints = np.unique(breakpoints)
            if len(breakpoints) < 2:
                return 0.0
                
            # Calculate frequencies
            expected_percents = np.histogram(expected, breakpoints)[0] / len(expected)
            actual_percents = np.histogram(actual, breakpoints)[0] / len(actual)
            
            # Avoid division by zero
            expected_percents = np.where(expected_percents == 0, 0.0001, expected_percents)
            actual_percents = np.where(actual_percents == 0, 0.0001, actual_percents)
            
            psi_value = np.sum((actual_percents - expected_percents) * np.log(actual_percents / expected_percents))
            return float(psi_value)
            
        except Exception:
            return 0.0

    def _calculate_kl(self, reference: np.ndarray, current: np.ndarray, buckets: int = 10) -> float:
        """
        Calculates KL Divergence (Current || Reference).
        """
        try:
            if len(reference) == 0 or len(current) == 0:
                return 0.0
            if np.min(reference) == np.max(reference):
                return 0.0
                
            # Use reference percentiles for binning (same as PSI)
            breakpoints = np.arange(0, buckets + 1) / (buckets) * 100
            breakpoints = np.percentile(reference, breakpoints)
            breakpoints = np.unique(breakpoints)
            
            if len(breakpoints) < 2:
                return 0.0
                
            ref_percents = np.histogram(reference, breakpoints)[0] / len(reference)
            curr_percents = np.histogram(current, breakpoints)[0] / len(current)
            
            # Smooth to avoid infinity
            ref_percents = np.where(ref_percents == 0, 0.0001, ref_percents)
            curr_percents = np.where(curr_percents == 0, 0.0001, curr_percents)
            
            return float(entropy(curr_percents, ref_percents))
        except Exception:
            return 0.0



import string
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import pandas as pd

from ..utils import pack_pipeline_output, resolve_columns, unpack_pipeline_input
from .base import BaseApplier, BaseCalculator
from ..registry import NodeRegistry
from ..core.meta.decorators import node_meta
from ..engines import SkyulfDataFrame, get_engine

# --- Constants ---
ALIAS_PUNCTUATION_TABLE = str.maketrans("", "", string.punctuation)
COMMON_BOOLEAN_ALIASES: Dict[str, str] = {
    "y": "Yes",
    "yes": "Yes",
    "true": "Yes",
    "1": "Yes",
    "on": "Yes",
    "t": "Yes",
    "affirmative": "Yes",
    "n": "No",
    "no": "No",
    "false": "No",
    "0": "No",
    "off": "No",
    "f": "No",
    "negative": "No",
}
COUNTRY_ALIAS_MAP: Dict[str, str] = {
    "usa": "USA",
    "us": "USA",
    "unitedstates": "USA",
    "unitedstatesofamerica": "USA",
    "states": "USA",
    "america": "USA",
    "unitedkingdom": "United Kingdom",
    "uk": "United Kingdom",
    "greatbritain": "United Kingdom",
    "england": "United Kingdom",
    "uae": "United Arab Emirates",
    "unitedarabemirates": "United Arab Emirates",
    "prc": "China",
    "peoplesrepublicofchina": "China",
    "southkorea": "South Korea",
    "republicofkorea": "South Korea",
    "sk": "South Korea",
}
TWO_DIGIT_YEAR_PIVOT = 50

# --- Helpers ---


def _auto_detect_text_columns(df: Union[pd.DataFrame, SkyulfDataFrame]) -> List[str]:
    engine = get_engine(df)
    if engine.name == "polars":
        import polars as pl

        return [
            c
            for c, t in zip(df.columns, df.dtypes)
            if t in [pl.Utf8, pl.Categorical, pl.Object]
        ]
    return list(df.select_dtypes(include=["object", "string", "category"]).columns)


def _auto_detect_numeric_columns(df: Union[pd.DataFrame, SkyulfDataFrame]) -> List[str]:
    engine = get_engine(df)
    if engine.name == "polars":
        import polars as pl

        return [
            c
            for c, t in zip(df.columns, df.dtypes)
            if t in [pl.Float32, pl.Float64, pl.Int8, pl.Int16, pl.Int32, pl.Int64, pl.UInt8, pl.UInt16, pl.UInt32, pl.UInt64]
        ]
    return list(df.select_dtypes(include=["number"]).columns)


def _auto_detect_datetime_columns(df: Union[pd.DataFrame, SkyulfDataFrame]) -> List[str]:
    engine = get_engine(df)
    if engine.name == "polars":
        import polars as pl

        return [
            c
            for c, t in zip(df.columns, df.dtypes)
            if t in [pl.Date, pl.Datetime] or isinstance(t, pl.Datetime)
        ]
    return list(df.select_dtypes(include=["datetime", "datetimetz"]).columns)


# --- Text Cleaning ---


class TextCleaningApplier(BaseApplier):
    def apply(  # noqa: C901
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        if not params or not params.get("columns"):
            return pack_pipeline_output(X, y, is_tuple)

        cols = params["columns"]
        operations = params.get("operations", [])

        valid_cols = [c for c in cols if c in X.columns]
        if not valid_cols or not operations:
            return pack_pipeline_output(X, y, is_tuple)

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            
            exprs = []
            for col in valid_cols:
                expr = pl.col(col).cast(pl.String)
                
                for op in operations:
                    op_type = op.get("op")
                    
                    if op_type == "trim":
                        mode = op.get("mode", "both")
                        if mode == "leading":
                            expr = expr.str.strip_chars_start()
                        elif mode == "trailing":
                            expr = expr.str.strip_chars_end()
                        else:
                            expr = expr.str.strip_chars()
                            
                    elif op_type == "case":
                        mode = op.get("mode", "lower")
                        if mode == "upper":
                            expr = expr.str.to_uppercase()
                        elif mode == "title":
                            expr = expr.str.to_titlecase()
                        elif mode == "sentence":
                            # Capitalize first char, lower the rest
                            expr = (
                                expr.str.slice(0, 1).str.to_uppercase() + 
                                expr.str.slice(1).str.to_lowercase()
                            )
                        else:
                            expr = expr.str.to_lowercase()
                            
                    elif op_type == "remove_special":
                        mode = op.get("mode", "keep_alphanumeric")
                        replacement = op.get("replacement", "")
                        
                        pattern = ""
                        if mode == "keep_alphanumeric":
                            pattern = r"[^a-zA-Z0-9]"
                        elif mode == "keep_alphanumeric_space":
                            pattern = r"[^a-zA-Z0-9\s]"
                        elif mode == "letters_only":
                            pattern = r"[^a-zA-Z]"
                        elif mode == "digits_only":
                            pattern = r"[^0-9]"
                            
                        if pattern:
                            expr = expr.str.replace_all(pattern, replacement)
                            
                    elif op_type == "regex":
                        mode = op.get("mode", "custom")
                        
                        if mode == "collapse_whitespace":
                            expr = expr.str.replace_all(r"\s+", " ").str.strip_chars()
                        elif mode == "extract_digits":
                            # extract first group
                            expr = expr.str.extract(r"(\d+)", 1)
                        elif mode == "normalize_slash_dates":
                            pass
                        elif mode == "custom":
                            pattern = op.get("pattern")
                            repl = op.get("repl", "")
                            if pattern:
                                expr = expr.str.replace_all(pattern, repl)
                                
                exprs.append(expr.alias(col))
            
            X_out = X.with_columns(exprs)
            return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        df_out = X.copy()

        for col in valid_cols:
            # Ensure column is string type
            if not pd.api.types.is_string_dtype(df_out[col]):
                df_out[col] = df_out[col].astype(str)

            series = df_out[col]

            for op in operations:
                op_type = op.get("op")

                if op_type == "trim":
                    mode = op.get("mode", "both")
                    if mode == "leading":
                        series = series.str.lstrip()
                    elif mode == "trailing":
                        series = series.str.rstrip()
                    else:
                        series = series.str.strip()

                elif op_type == "case":
                    mode = op.get("mode", "lower")
                    if mode == "upper":
                        series = series.str.upper()
                    elif mode == "title":
                        series = series.str.title()
                    elif mode == "sentence":
                        series = series.str.capitalize()
                    else:
                        series = series.str.lower()

                elif op_type == "remove_special":
                    mode = op.get("mode", "keep_alphanumeric")
                    replacement = op.get("replacement", "")

                    if mode == "keep_alphanumeric":
                        # Keep letters and numbers
                        series = series.str.replace(
                            r"[^a-zA-Z0-9]", replacement, regex=True
                        )
                    elif mode == "keep_alphanumeric_space":
                        # Keep letters, numbers, and spaces
                        series = series.str.replace(
                            r"[^a-zA-Z0-9\s]", replacement, regex=True
                        )
                    elif mode == "letters_only":
                        series = series.str.replace(
                            r"[^a-zA-Z]", replacement, regex=True
                        )
                    elif mode == "digits_only":
                        series = series.str.replace(r"[^0-9]", replacement, regex=True)

                elif op_type == "regex":
                    mode = op.get("mode", "custom")

                    if mode == "collapse_whitespace":
                        series = series.str.replace(r"\s+", " ", regex=True).str.strip()
                    elif mode == "extract_digits":
                        series = series.str.extract(r"(\d+)", expand=False)
                    elif mode == "normalize_slash_dates":
                        # Simple attempt to fix 1/1/2020 -> 01/01/2020
                        # This is complex, maybe just a placeholder
                        pass
                    elif mode == "custom":
                        pattern = op.get("pattern")
                        repl = op.get("repl", "")
                        if pattern:
                            series = series.str.replace(pattern, repl, regex=True)

            df_out[col] = series

        return pack_pipeline_output(df_out, y, is_tuple)


@NodeRegistry.register("TextCleaning", TextCleaningApplier)
@node_meta(
    id="TextCleaning",
    name="Text Cleaning",
    category="Cleaning",
    description="Clean text data (trim, case conversion, remove special chars).",
    params={"columns": [], "operations": []}
)
class TextCleaningCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        # Config:
        # columns: List[str]
        # operations: List[Dict]
        #   {'op': 'trim', 'mode': 'both'|'leading'|'trailing'}
        #   {'op': 'case', 'mode': 'lower'|'upper'|'title'|'sentence'}
        #   {'op': 'remove_special', 'mode': 'keep_alphanumeric'|'keep_alphanumeric_space'|'letters_only'|'digits_only',
        #    'replacement': ''}
        #   {'op': 'regex', 'mode': 'custom'|'collapse_whitespace'|'extract_digits'|'normalize_slash_dates',
        #    'pattern': '...', 'repl': '...'}

        X, _, _ = unpack_pipeline_input(df)

        cols = resolve_columns(X, config, _auto_detect_text_columns)

        if not cols:
            return {}

        operations = config.get("operations", [])

        return {"type": "text_cleaning", "columns": cols, "operations": operations}


# --- Invalid Value Replacement ---


class InvalidValueReplacementApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        cols = params.get("columns", [])
        
        # Params
        replace_inf = params.get("replace_inf", False)
        replace_neg_inf = params.get("replace_neg_inf", False)
        rule = params.get("rule")
        
        # Unify replacement value
        replacement = params.get("replacement", np.nan)
        value = params.get("value")
        final_replacement = value if value is not None else replacement
        
        min_value = params.get("min_value")
        max_value = params.get("max_value")

        valid_cols = [c for c in cols if c in X.columns]
        if not valid_cols:
            return pack_pipeline_output(X, y, is_tuple)

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            
            exprs = []
            for col in valid_cols:
                expr = pl.col(col)
                
                # 1. Handle Inf
                if replace_inf:
                    expr = pl.when(expr == float('inf')).then(final_replacement).otherwise(expr)
                if replace_neg_inf:
                    expr = pl.when(expr == float('-inf')).then(final_replacement).otherwise(expr)
                    
                # 2. Handle Rules
                if rule == "negative":
                    expr = pl.when(expr < 0).then(final_replacement).otherwise(expr)
                elif rule == "zero":
                    expr = pl.when(expr == 0).then(final_replacement).otherwise(expr)
                elif rule == "custom_range":
                    if min_value is not None and max_value is not None:
                        expr = pl.when((expr < min_value) | (expr > max_value)).then(final_replacement).otherwise(expr)
                    elif min_value is not None:
                        expr = pl.when(expr < min_value).then(final_replacement).otherwise(expr)
                    elif max_value is not None:
                        expr = pl.when(expr > max_value).then(final_replacement).otherwise(expr)
                        
                exprs.append(expr.alias(col))
            
            X_out = X.with_columns(exprs)
            return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        df_out = X.copy()
        
        for col in valid_cols:
            # Ensure numeric
            df_out[col] = pd.to_numeric(df_out[col], errors="coerce")
            
            # 1. Handle Inf
            to_replace = []
            if replace_inf:
                to_replace.append(np.inf)
            if replace_neg_inf:
                to_replace.append(-np.inf)
            
            if to_replace:
                df_out[col] = df_out[col].replace(to_replace, final_replacement)
                
            # 2. Handle Rules
            series = df_out[col]
            mask = None
            
            if rule == "negative":
                mask = series < 0
            elif rule == "zero":
                mask = series == 0
            elif rule == "negative_to_nan":
                mask = series < 0
            elif rule == "custom_range":
                if min_value is not None and max_value is not None:
                    mask = (series < min_value) | (series > max_value)
                elif min_value is not None:
                    mask = series < min_value
                elif max_value is not None:
                    mask = series > max_value
            
            if mask is not None:
                df_out.loc[mask, col] = final_replacement

        return pack_pipeline_output(df_out, y, is_tuple)


@NodeRegistry.register("InvalidValueReplacement", InvalidValueReplacementApplier)
@node_meta(
    id="InvalidValueReplacement",
    name="Replace Invalid Values",
    category="Cleaning",
    description="Replace specified values with nan.",
    params={"columns": [], "invalid_values": []}
)
class InvalidValueReplacementCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        X, _, _ = unpack_pipeline_input(df)
        cols = resolve_columns(X, config, _auto_detect_numeric_columns)
        
        return {
            "type": "invalid_value_replacement",
            "columns": cols,
            "replace_inf": config.get("replace_inf", False),
            "replace_neg_inf": config.get("replace_neg_inf", False),
            "rule": config.get("rule") or config.get("mode"),
            "replacement": config.get("replacement", np.nan),
            "value": config.get("value"),
            "min_value": config.get("min_value"),
            "max_value": config.get("max_value"),
        }


class ValueReplacementApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        cols = params.get("columns", [])
        mapping = params.get("mapping")
        to_replace = params.get("to_replace")
        value = params.get("value")

        valid_cols = [c for c in cols if c in X.columns]
        if not valid_cols:
            return pack_pipeline_output(X, y, is_tuple)

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            
            exprs = []
            
            if mapping:
                is_nested = any(isinstance(v, dict) for v in mapping.values())
                
                if is_nested:
                    # Column-specific mapping
                    for col in valid_cols:
                        if col in mapping:
                            map_dict = mapping[col]
                            # Polars replace: mapping dict
                            # We need to ensure types match or use strict=False
                            # map_dict keys/values should be compatible with col type
                            exprs.append(pl.col(col).replace(map_dict, default=pl.col(col)).alias(col))
                else:
                    # Global mapping
                    for col in valid_cols:
                        exprs.append(pl.col(col).replace(mapping, default=pl.col(col)).alias(col))
            
            else:
                # Global replacement
                if to_replace is not None:
                    for col in valid_cols:
                        if isinstance(to_replace, (dict, pd.Series)) or hasattr(to_replace, "items"):
                             exprs.append(pl.col(col).replace(to_replace, default=pl.col(col)).alias(col))
                        else:
                            # Replace single value
                            # pl.when(col == to_replace).then(value).otherwise(col)
                            # But replace is easier
                            exprs.append(pl.col(col).replace({to_replace: value}, default=pl.col(col)).alias(col))
            
            if exprs:
                X_out = X.with_columns(exprs)
                return pack_pipeline_output(X_out, y, is_tuple)
            return pack_pipeline_output(X, y, is_tuple)

        # Pandas Path
        df_out = X.copy()

        if mapping:
            # Check if mapping is column-specific (nested dict) or global (simple dict)
            is_nested = any(isinstance(v, dict) for v in mapping.values())

            if is_nested:
                # Column-specific mapping
                for col, map_dict in mapping.items():
                    if col in valid_cols:
                        df_out[col] = df_out[col].replace(map_dict)
            else:
                # Global mapping applied to all valid columns
                for col in valid_cols:
                    df_out[col] = df_out[col].replace(mapping)

        else:
            # Global replacement
            if to_replace is not None:
                for col in valid_cols:
                    # Check for dict-like objects (dict, pd.Series, etc.)
                    # If to_replace is a mapping, we ignore `value` and apply the mapping.
                    if isinstance(to_replace, (dict, pd.Series)) or hasattr(to_replace, "items"):
                        df_out[col] = df_out[col].replace(to_replace)
                    else:
                        df_out[col] = df_out[col].replace(to_replace, value)

        return pack_pipeline_output(df_out, y, is_tuple)


@NodeRegistry.register("ValueReplacement", ValueReplacementApplier)
@node_meta(
    id="ValueReplacement",
    name="Replace Values",
    category="Cleaning",
    description="Replace specified values with new values.",
    params={"columns": [], "mapping": {}}
)
class ValueReplacementCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        X, _, _ = unpack_pipeline_input(df)

        # Config: {'columns': [...], 'mapping': {col: {old: new}}}
        # OR {'columns': [...], 'to_replace': '?', 'value': np.nan}
        # OR {'columns': [...], 'replacements': [{'old': 1, 'new': 100}, ...]}

        cols = resolve_columns(X, config)
        mapping = config.get("mapping")
        to_replace = config.get("to_replace")
        value = config.get("value")
        replacements = config.get("replacements")

        if replacements:
            # Convert list of dicts to mapping dict
            mapping = {item["old"]: item["new"] for item in replacements}

        return {
            "type": "value_replacement",
            "columns": cols,
            "mapping": mapping,
            "to_replace": to_replace,
            "value": value,
        }


# --- Alias Replacement ---


class AliasReplacementApplier(BaseApplier):
    def apply(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        params: Dict[str, Any],
    ) -> Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...]]:
        X, y, is_tuple = unpack_pipeline_input(df)
        engine = get_engine(X)

        cols = params.get("columns", [])
        alias_type = params.get("alias_type", "boolean")
        custom_map = params.get("custom_map", {})

        valid_cols = [c for c in cols if c in X.columns]
        if not valid_cols:
            return pack_pipeline_output(X, y, is_tuple)

        # Prepare the mapping
        mapping = {}
        if alias_type == "boolean":
            mapping = COMMON_BOOLEAN_ALIASES
        elif alias_type == "country":
            mapping = COUNTRY_ALIAS_MAP
        elif alias_type == "custom":
            mapping = custom_map

        # Polars Path
        if engine.name == "polars":
            import polars as pl
            import re
            
            # Construct regex for punctuation removal
            # We want to remove punctuation and spaces, then lower case
            # Python: val.lower().translate(ALIAS_PUNCTUATION_TABLE).replace(" ", "")
            
            escaped_punct = re.escape(string.punctuation)
            
            exprs = []
            for col in valid_cols:
                # 1. Clean: lower -> remove punct -> remove space
                clean_expr = (
                    pl.col(col)
                    .cast(pl.String)
                    .str.to_lowercase()
                    .str.replace_all(f"[{escaped_punct}]", "")
                    .str.replace_all(" ", "")
                )
                
                # 2. Map
                # replace(mapping, default=None) -> if not found, returns null? 
                # No, replace returns default if not found.
                # We want: if found in mapping, use mapped value. If not found, use ORIGINAL value.
                # So:
                # mapped = clean.replace(mapping, default=None)
                # result = coalesce(mapped, original)
                
                # Note: replace in Polars with a dict maps keys to values.
                # If a value is not in keys, it keeps it AS IS (if default is not set) or uses default.
                # But here 'clean_expr' is the CLEANED value (e.g. "usa").
                # If "usa" is in mapping, it becomes "USA".
                # If "foobar" is NOT in mapping, it stays "foobar".
                # But we want to revert to ORIGINAL "Foo Bar" if not found?
                # The pandas code says: mapped_series.fillna(df_out[col])
                # Pandas map returns NaN if not found.
                # Polars replace returns original value if not found (by default).
                # So we need `replace(mapping, default=None)` to get nulls for non-matches.
                
                mapped_expr = clean_expr.replace(mapping, default=None)
                
                # 3. Coalesce with original
                final_expr = pl.coalesce([mapped_expr, pl.col(col)])
                
                exprs.append(final_expr.alias(col))
            
            X_out = X.with_columns(exprs)
            return pack_pipeline_output(X_out, y, is_tuple)

        # Pandas Path
        df_out = X.copy()

        # Normalize function
        def normalize(val):
            if not isinstance(val, str):
                return val
            # Remove punctuation and lower case for matching
            clean = val.lower().translate(ALIAS_PUNCTUATION_TABLE).replace(" ", "")
            return mapping.get(clean, val)

        for col in valid_cols:
            # Apply normalization
            # This is slow for large dataframes, vectorization is hard with custom fuzzy logic
            # We use map for exact matches on cleaned keys

            # 1. Clean the column temporarily
            clean_series = (
                df_out[col]
                .astype(str)
                .str.lower()
                .str.translate(ALIAS_PUNCTUATION_TABLE)
                .str.replace(" ", "")
            )

            # 2. Map values
            mapped_series = clean_series.map(mapping)

            # 3. Fill NaNs (unmapped) with original values
            df_out[col] = mapped_series.fillna(df_out[col])

        return pack_pipeline_output(df_out, y, is_tuple)


@NodeRegistry.register("AliasReplacement", AliasReplacementApplier)
@node_meta(
    id="AliasReplacement",
    name="Standardize Values",
    category="Cleaning",
    description="Standardize common variations in text values (e.g. Yes/No, Country names).",
    params={"columns": [], "domain": "boolean"}
)
class AliasReplacementCalculator(BaseCalculator):
    def fit(
        self,
        df: Union[pd.DataFrame, SkyulfDataFrame, Tuple[Any, ...], Any],
        config: Dict[str, Any],
    ) -> Dict[str, Any]:
        X, _, _ = unpack_pipeline_input(df)
        cols = resolve_columns(X, config, _auto_detect_text_columns)

        # Support both 'alias_type' and 'mode'
        alias_type = config.get("alias_type") or config.get("mode", "boolean")

        # Map legacy modes
        if alias_type == "normalize_boolean":
            alias_type = "boolean"
        elif alias_type == "canonicalize_country_codes":
            alias_type = "country"

        # Support both 'custom_map' and 'custom_pairs'
        custom_map = config.get("custom_map") or config.get("custom_pairs", {})

        # Normalize custom_map keys to match the cleaning logic in Applier
        if custom_map:
            normalized_map = {}
            for k, v in custom_map.items():
                if isinstance(k, str):
                    clean_k = (
                        k.lower().translate(ALIAS_PUNCTUATION_TABLE).replace(" ", "")
                    )
                    normalized_map[clean_k] = v
                else:
                    normalized_map[k] = v
            custom_map = normalized_map

        return {
            "type": "alias_replacement",
            "columns": cols,
            "alias_type": alias_type,
            "custom_map": custom_map,
        }



